# Copyright © 2021. TIBCO Software Inc.
# This file is subject to the license terms contained
# in the license file that is distributed with this file.

"""Pure Python implementation of a Spotfire package (.spk extension) builder tool."""

import abc
import argparse
import datetime
import io
import json
import os
import platform
import re
import shutil
import subprocess
import sys
import tempfile
import time
import typing
import uuid
from xml.etree import ElementTree
import zipfile

from pip._vendor.packaging import version as pip_version

import spotfire
import spotfire.version

# Command line parsing helpers

CLI_PARSER = argparse.ArgumentParser(prog="python -m spotfire.spk")
CLI_SUBPARSERS = CLI_PARSER.add_subparsers(dest="subcommand")


def subcommand(args=None, parent=CLI_SUBPARSERS):
    """Decorate a function so that it is usable as a subcommand of 'python -m spotfire.spk'."""
    if args is None:
        args = []

    def decorator(func):
        """Decorate the function."""
        parser = parent.add_parser(func.__name__, help=func.__doc__)
        for arg in args:
            parser.add_argument(*arg[0], **arg[1])
        parser.set_defaults(func=func)
        return func

    return decorator


def argument(*name_or_flags, **kwargs):
    """Used by the '@subcommand()' decorator to add arguments to the created subcommand."""
    return list(name_or_flags), kwargs


# Messaging helper functions

def _message(msg: str) -> None:
    sys.stdout.write(msg)
    sys.stdout.write(os.linesep)


def _error(msg: str) -> None:
    sys.stdout.write(msg)
    sys.stdout.write(os.linesep)


# Packaging helper functions

class _SpkVersion:
    """Represents a version number as presented in Spotfire SPK package files.  Version numbers have four components
    (organized from largest scope to smallest): major, minor, service pack, and identifier."""

    def __init__(self, major: int = 1, minor: int = 0, service_pack: int = 0, identifier: int = 0) -> None:
        self._versions = [major, minor, service_pack, identifier]

    @staticmethod
    def from_str(str_: str) -> '_SpkVersion':
        """Parse a string into a version.

        :param str_: the string to parse into a version number
        """
        components = str_.split(".")
        if len(components) > 4:
            raise ValueError("Must have four components in version number.")
        return _SpkVersion(*map(int, components))

    @staticmethod
    def from_version_info(identifier: int) -> '_SpkVersion':
        """Extract the SPK version of the running Python installation."""
        spk_minor = (sys.version_info.minor * 100) + sys.version_info.micro
        spotfire_version_components = spotfire.version.__version__.split(".")
        spk_service_pack = (int(spotfire_version_components[0]) * 10000) + \
                           (int(spotfire_version_components[1]) * 100) + \
                           (int(spotfire_version_components[2]))
        version = _SpkVersion(sys.version_info.major, spk_minor, spk_service_pack, identifier)
        return version

    def __str__(self):
        # pylint: disable=consider-using-f-string
        return "%d.%d.%d.%d" % tuple(self._versions)

    def __repr__(self):
        return f"{self.__class__.__module__}.{self.__class__.__qualname__}{tuple(self._versions)!r}"

    def increment_major(self) -> None:
        """Increment the major component of the version number.  Resets all smaller components to zero."""
        self._versions[0] += 1
        self._versions[1:] = [0, 0, 0]

    def increment_minor(self) -> None:
        """Increment the minor component of the version number.  Resets all smaller components to zero."""
        self._versions[1] += 1
        self._versions[2:] = [0, 0]

    def _decrement(self, pos, wrap_around) -> None:
        while pos >= 0:
            self._versions[pos] -= 1
            if self._versions[pos] < 0:
                self._versions[pos] = wrap_around
                pos -= 1
            else:
                return
        self._versions = [0, 0, 0, 0]
        raise ValueError("Version object cannot decrement major version below zero.")

    def decrement_major(self) -> None:
        """Decrement the major component of the version number.  Resets all smaller components to zero.

        :raises ValueError: if the major component would be decremented below zero
        """
        self._versions[1:] = [0, 0, 0]
        self._decrement(0, 0)

    def decrement_minor(self, wrap_around: int = 99) -> None:
        """Decrement the minor component of the version number.  Resets all smaller components to zero.

        :param wrap_around: the value to set components that are decremented below zero
        """
        self._versions[2:] = [0, 0]
        self._decrement(1, wrap_around)

    def decrement_service_pack(self, wrap_around: int = 99) -> None:
        """Decrement the service pack component of the version number.  Resets all smaller components to zero.

        :param wrap_around: the value to set components that are decremented below zero
        """
        self._versions[3:] = [0]
        self._decrement(2, wrap_around)

    def __lt__(self, other):
        if not isinstance(other, _SpkVersion):
            return NotImplemented
        return self._versions < other._versions  # pylint: disable=protected-access


def _tee(command: typing.List[str], encoding: str = "utf-8") -> typing.List[str]:
    """Run an external command, capturing the output and displaying it to standard output.  Standard error is collapsed
    into the same stream as standard output.

    :param command: the command to run
    :param encoding: the encoding to use to convert the bytes of the command's output into str
    :return: a list of the lines of the command's output
    :raises CalledProcessError: if the command returns a non-zero exit status
    """
    output = []
    sys.stdout.flush()
    with subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT) as process:
        for line in process.stdout:
            line = line.decode(encoding)
            output.append(line)
            sys.stdout.write(line)
        return_code = process.wait()
        if return_code:
            raise subprocess.CalledProcessError(return_code, command[0])
        return output


def _brand_file(filename: str, data: typing.Dict, comment: str, line_length: int = 72) -> None:
    """Brand a file with the JSON representation of data.

    :param filename: the filename of the file to brand
    :param data: the data to brand into the file
    :param comment: the comment characters to prefix each line of the brand
    :param line_length: the number of characters that should be present on each line of the brand
    """
    data_json = json.dumps(data, separators=(',', ':'))

    # Read in the file, ignoring any current brand.
    lines = []
    with open(filename, "r", encoding="utf8") as file:
        for line in file.readlines():
            if not line.startswith(comment):
                lines.append(line)

    # Add line break to ensure brand is on a new line (PYSRV-162)
    if not lines[-1].endswith("\n"):
        lines[-1] += "\n"

    # Now append the brand lines to what we just read in.
    brand_per_line = line_length - len(comment)
    while data_json:
        lines.append(f"{comment}{data_json[:brand_per_line]}\n")
        data_json = data_json[brand_per_line:]

    # Write out the new file.
    with open(filename, "w", encoding="utf8") as file:
        file.writelines(lines)


def _read_brand(filename: str, comment: str) -> typing.Dict:
    """Read a brand from a file.

    :param filename: the filename of the file with the brand
    :param comment: the comment characters that prefix each line of the brand
    :return: the data that is encoded in the brand
    """
    # Read in the file, pulling in only the lines with branding on them.
    lines = []
    with open(filename, "r", encoding="utf8") as file:
        for line in file.readlines():
            if line.startswith(comment):
                lines.append(line)

    # Reassemble data and parse the JSON object from the brand.
    data_json = "".join(map(lambda x: x[len(comment):].rstrip(), lines))
    return json.loads(data_json) if data_json else {}


class _PackageBuilder(metaclass=abc.ABCMeta):
    # pylint: disable=too-many-instance-attributes,consider-using-f-string
    def __init__(self) -> None:
        self.name = None
        self.version = None
        self.id = None
        self.output = None
        self.excludes = []
        self.last_scan_dir = None
        self._contents = []
        if platform.system() == "Windows":
            self._site_packages_dirname = "Lib\\site-packages"
        else:
            self._site_packages_dirname = "lib/python%d.%d/site-packages" % sys.version_info[:2]

    def add(self, filename: str, archive_name: str) -> None:
        """Add a file to the package."""
        archive_backslash = archive_name.replace("/", "\\")
        if self.excludes is not None:
            for exclude in self.excludes:
                if archive_backslash.startswith(exclude):
                    return
        self._contents.append((filename, archive_backslash))

    def scan_python_installation(self, prefix: str) -> None:
        """Scan all the files in the Python installation."""
        try:
            py_prefix = sys.base_prefix
        except AttributeError:
            py_prefix = getattr(sys, "real_prefix", sys.prefix)
        _message(f"Scanning Python installation at {py_prefix} for files to include.")

        # Scan all files in the running Python's prefix
        for root, _, filenames in os.walk(py_prefix):
            for filename in filenames:
                filename_ondisk = os.path.join(root, filename)
                filename_relative = os.path.relpath(filename_ondisk, py_prefix)
                filename_payload = f"{prefix}/{filename_relative}"
                if not filename_relative.startswith(self._site_packages_dirname):
                    # Omit any packages installed in site-packages
                    self.add(filename_ondisk, filename_payload)

    def scan_spotfire_package(self, prefix: str) -> None:
        """Scan the 'spotfire' package (and it's .dist-info directory) into site-packages."""
        _message("Scanning 'spotfire' package files.")

        # Grab the files of the 'spotfire' package.
        for root, _, filenames in os.walk(spotfire.__path__[0]):
            for filename in filenames:
                filename_ondisk = os.path.join(root, filename)
                filename_relative = os.path.relpath(filename_ondisk, spotfire.__path__[0])
                filename_payload = f"{prefix}/{self._site_packages_dirname}/spotfire/{filename_relative}"
                self.add(filename_ondisk, filename_payload)

        # Grab the .dist-info directory.
        site_packages_dir = os.path.dirname(spotfire.__path__[0])
        spotfire_dist_info_re = re.compile(r"^spotfire(-.*)?\.(dist|egg)-info$")
        for subdir in next(os.walk(site_packages_dir))[1]:
            if spotfire_dist_info_re.match(subdir):
                spotfire_dist_info_dir = os.path.join(site_packages_dir, subdir)
                for root, _, filenames in os.walk(spotfire_dist_info_dir):
                    for filename in filenames:
                        filename_ondisk = os.path.join(root, filename)
                        filename_relative = os.path.relpath(filename_ondisk, spotfire_dist_info_dir)
                        filename_payload = f"{prefix}/{self._site_packages_dirname}/{subdir}/{filename_relative}"
                        self.add(filename_ondisk, filename_payload)

    def scan_requirements_txt(self, requirements: str, constraint: str, prefix: str, prefix_direct: bool = False,
                              use_deny_list: bool = False) -> \
            typing.Tuple[typing.Callable[[], None], typing.Dict[str, str]]:
        """Scan the contents of a pip 'requirements.txt' file into site-packages.

        :param requirements: the filename of the requirements file that declares the pip packages to put in the
        SPK package
        :param constraint: the filename of the constraints file that declares the constraints pip should apply
        when resolving requirements
        :param prefix: the directory prefix under which the pip packages should be scanned into
        :param prefix_direct: whether to scan the pip packages into the ``site-packages`` directory
        :param use_deny_list: whether to delete the packages included with the Interpreter from the
                                temporary package area before bundling.
        :return: a function that cleans up the temporarily installed packages, and a dict that maps the names of
        pip packages that were scanned into the SPK package to their installed versions
        """
        tempdir = tempfile.mkdtemp(prefix="spk")
        self.last_scan_dir = tempdir

        def cleanup():
            """Cleanup the created tempdir."""
            shutil.rmtree(tempdir)

        # Install the packages from the requirement file into tempdir.
        _message(f"Installing pip packages from {requirements} to temporary location.")
        try:
            command = [sys.executable, "-m", "pip", "install", "--upgrade", "--disable-pip-version-check",
                             "--target", tempdir, "--requirement", requirements]
            if constraint:
                command.extend(["--constraint", constraint])
            pip_output = _tee(command)
        except subprocess.CalledProcessError:
            _error("Error installing required packages.  Aborting.")
            cleanup()
            sys.exit(1)

        # Update the brand with the package versions.
        package_versions = self._packages_installed(pip_output)

        # Delete duplicate packages if needed
        if use_deny_list:
            package_versions = self.scan_duplicate_packages(tempdir, package_versions)

        # Scan all files in tempdir
        _message("Scanning package files from temporary location.")
        for root, _, filenames in os.walk(tempdir):
            for filename in filenames:
                filename_ondisk = os.path.join(root, filename)
                filename_relative = os.path.relpath(filename_ondisk, tempdir)
                if prefix_direct:
                    filename_payload = f"{prefix}/{filename_relative}"
                else:
                    filename_payload = f"{prefix}/{self._site_packages_dirname}/{filename_relative}"
                if not filename_relative.startswith(f"bin{os.path.sep}"):
                    self.add(filename_ondisk, filename_payload)

        return cleanup, package_versions

    @staticmethod
    def scan_duplicate_packages(tempdir: str, package_versions):
        """ When building a custom Python Packages SPK, find and delete duplicate packages.
        :return:
        """
        # Disable Pylint `Too many nested blocks`
        # pylint: disable=too-many-nested-blocks,too-many-locals
        # Use the spotfire requirements file as a deny list.
        _message(f"{spotfire.__path__[0]}")
        if "spotfire.zip" in spotfire.__path__[0]:
            # Handle getting the requirements.txt from a zip file.
            with zipfile.ZipFile(os.path.split(spotfire.__path__[0])[0]) as zfile:
                with zfile.open("spotfire/requirements.txt") as deny_file:
                    deny_requirements = deny_file.readlines()
                deny_list = [line.decode("utf-8").rstrip('\n') for line in deny_requirements]
        else:
            # Handle getting the requirements.txt directly from the file system.
            with open(os.path.join(spotfire.__path__[0], "requirements.txt"), encoding="utf8") as deny_file:
                deny_requirements = deny_file.readlines()
            deny_list = [line.rstrip('\n') for line in deny_requirements]

        # Compile a Set of directories to delete
        package_directories = set()
        # Go through each package
        for package_name in deny_list:
            # Remove version numbers from package name and keep a copy of the original name
            package_name = package_name.split(' ', 1)[0]
            original_package_name = package_name
            # Convert dash to underscore for file access
            package_name = package_name.replace('-', '_')
            # Walk through each directory
            for root, directory_names, _ in os.walk(tempdir):
                for directory_name in directory_names:
                    # look for the `dist-info` directory for each package
                    if directory_name.startswith(package_name) and directory_name.endswith('dist-info'):
                        package_directories.add(directory_name)
                        # Go through the RECORD file from the package dist-info and delete each listed file
                        record_file = os.path.join(root, directory_name, 'RECORD')
                        if os.path.isfile(record_file):
                            with open(record_file, encoding="utf8") as open_record_file:
                                record_files = open_record_file.readlines()
                            for file in record_files:
                                # Clean up the path to each file.
                                # RECORD has file name, then comma followed by info we don't need.
                                file = file.split(',', 1)[0]
                                # Files in "/bin" start with "../../" but we don't need that.
                                if file.startswith('../../'):
                                    file = file.split('../../', 1)[1]
                                # Get the path to the file
                                file = os.path.join(root, file)
                                # Delete the file if it exists
                                if os.path.isfile(file):
                                    os.remove(file)
                                else:
                                    _message(f"Could not find RECORD file {file}")
                            _message(f"Deleted files listed in RECORD for {directory_name}")
                            package_versions.pop(original_package_name, None)
                    if directory_name == package_name:
                        package_directories.add(directory_name)
        # Manually add 'dateutil's non-standard package name
        package_directories.add('dateutil')
        # Delete the packages' directories
        for package_directory in package_directories:
            package_directory_file = os.path.join(tempdir, package_directory)
            if os.path.isdir(package_directory_file):
                try:
                    shutil.rmtree(package_directory_file)
                except OSError:
                    _message(f"Unable to remove directory {package_directory_file}")
        _message("Completed removing files and directories for packages on the deny list.")
        return package_versions

    @staticmethod
    def _packages_installed(pip_output):
        """Determine what packages were installed by parsing pip's output."""
        _installed = "Successfully installed "
        # Scan the output for the line with the magic string
        installed_line = None
        for i in reversed(pip_output):
            if i.startswith(_installed):
                installed_line = i
        # Now parse the line to get the package versions
        package_versions = {}
        if installed_line:
            installed_packages = installed_line[len(_installed):]
            for pkg in installed_packages.strip().split(" "):
                pkg_dash_pos = pkg.rfind("-")
                package_versions[pkg[:pkg_dash_pos]] = pkg[pkg_dash_pos + 1:]
        return package_versions

    @abc.abstractmethod
    def _payload_name(self) -> str:
        """Get the payload archive name for this package."""

    def _create_module(self) -> ElementTree.Element:
        """Create the module document."""
        module = ElementTree.Element("module")
        module_id = ElementTree.SubElement(module, "id")
        module_id.text = self.id
        module_name = ElementTree.SubElement(module, "name")
        module_name.text = self.name
        module_version = ElementTree.SubElement(module, "version")
        module_version.text = str(self.version)
        return module

    def _create_metadata(self, module: ElementTree.Element) -> ElementTree.Element:
        """Create the metadata document."""
        metadata = ElementTree.Element("Package", {
            "SchemaVersion": "2.0",
            "Name": self.name,
            "Version": str(self.version),
            "LastModified": datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.0000000Z"),
            "SeriesId": self.id,
            "InstanceId": str(uuid.uuid4()),
            "CabinetName": self._payload_name(),
            "Type": "Module",
        })
        ElementTree.SubElement(metadata, "Files")
        ElementTree.SubElement(metadata, "Assemblies")
        metadata_client_module = ElementTree.SubElement(metadata, "ClientModule")
        for child in module:
            metadata_client_module.append(child)
        return metadata

    @abc.abstractmethod
    def _build_payload(self, metadata: ElementTree.Element, module: ElementTree.Element, payload_dest: str) -> None:
        """Build the main payload archive for the SPK package."""

    def build(self) -> None:
        """Build the SPK package."""

        # Create the module and metadata documents
        module = self._create_module()
        metadata = self._create_metadata(module)

        # Assemble things
        payload_fd, payload_tempfile = tempfile.mkstemp(prefix="spk")
        os.close(payload_fd)
        try:
            _message(f"Building Spotfire SPK package {self.output}.")

            # Assemble the payload zip
            self._build_payload(metadata, module, payload_tempfile)

            # Now assemble the SPK file
            with zipfile.ZipFile(self.output, "w", compression=zipfile.ZIP_DEFLATED) as spk:
                spk.writestr("module.xml", _et_to_bytes(module))
                spk.writestr("Metadata.xml", _et_to_bytes(metadata))
                spk.write(payload_tempfile, f"Contents/{self._payload_name()}")
            _message("Done.")
        finally:
            os.unlink(payload_tempfile)


def _et_to_bytes(element: ElementTree.Element) -> bytes:
    # Add indentation
    _et_indent(element)

    # Serialize the element to a string
    temp_io = io.BytesIO()
    ElementTree.ElementTree(element).write(temp_io, encoding="utf-8", xml_declaration=True)
    return temp_io.getvalue()


def _et_indent(element: ElementTree.Element, indent="  ", level=0) -> None:
    # pylint: disable=redefined-argument-from-local
    indent_text = "\n" + (level * indent)
    if element:
        if not element.text or not element.text.strip():
            element.text = indent_text + indent
        if not element.tail or not element.tail.strip():
            element.tail = indent_text
        for element in element:
            _et_indent(element, indent, level + 1)
        if not element.tail or not element.tail.strip():
            element.tail = indent_text
    else:
        if level and (not element.tail or not element.tail.strip()):
            element.tail = indent_text


class _ZipPackageBuilder(_PackageBuilder):
    # pylint: disable=too-many-instance-attributes

    def __init__(self):
        super().__init__()
        self.chmod_script_name = None

    def _payload_name(self) -> str:
        """Get the payload archive name for this package."""
        return f"{self.name}.zip"

    def _create_module(self) -> ElementTree.Element:
        """Create the module document."""
        module = super()._create_module()
        module_intended_client = ElementTree.SubElement(module, "intendedClient")
        module_intended_client.text = "PythonService"
        module_intended_platform = ElementTree.SubElement(module, "intendedPlatform")
        module_intended_platform.text = platform.system().upper()
        module_webplayer_folder = ElementTree.SubElement(module, "webPlayerContentFolder")
        module_webplayer_folder.text = "root"
        return module

    def _create_metadata(self, module: ElementTree.Element) -> ElementTree.Element:
        """Create the metadata document."""
        metadata = super()._create_metadata(module)
        metadata_archive = ElementTree.SubElement(metadata, "ArchiveFormat")
        metadata_archive.text = "zip"
        return metadata

    def _build_payload(self, metadata: ElementTree.Element, module: ElementTree.Element, payload_dest: str) -> None:
        """Build the main payload archive for the SPK package."""
        metadata_files = metadata.find("Files")
        payload_script = []
        with zipfile.ZipFile(payload_dest, "w", compression=zipfile.ZIP_DEFLATED) as payload:
            # Add all files that are supposed to go into the package
            for filename_ondisk, filename_payload in self._contents:
                filename_payload_fwdslash = filename_payload.replace("\\", "/")[5:]
                if os.path.islink(filename_ondisk):
                    payload_script += "if [ ! -e {payload} ]; then ln -s {ondisk} {payload}; fi\n".format(
                        payload=filename_payload_fwdslash,
                        ondisk=os.readlink(filename_ondisk))
                else:
                    payload.write(filename_ondisk, filename_payload.replace("\\", "/"))
                    stat = os.lstat(filename_ondisk)
                    mode = stat.st_mode & 0o7777
                    ElementTree.SubElement(metadata_files, "File", {
                        "TargetRelativePath": filename_payload,
                        "LastModifiedDate": datetime.datetime.utcfromtimestamp(stat.st_ctime).strftime(
                            "%Y-%m-%dT%H:%M:%SZ"),
                    })
                    if platform.system() != "Windows" and mode != 0o644:
                        payload_script += f"chmod {mode:o} {filename_payload_fwdslash}\n"

            # Add the payload script if we added any lines
            if payload_script:
                payload_script.insert(0, "#!/bin/sh\n")
                payload.writestr(f"root/Tools/Update/{self.chmod_script_name}.sh", "".join(payload_script))
                ElementTree.SubElement(metadata_files, "File", {
                    "TargetRelativePath": f"root\\Tools\\Update\\{self.chmod_script_name}",
                    "LastModifiedDate": datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
                })

            # Add the module.xml file
            payload.writestr("module.xml", _et_to_bytes(module))


class _CabPackageBuilder(_PackageBuilder):
    # pylint: disable=too-many-instance-attributes

    def __init__(self):
        if platform.system() != "Windows":
            _error("Cabinet based SPK packages cannot be built on non-Windows systems.  Aborting.")
            sys.exit(1)
        super().__init__()
        self.cert_file = None
        self.cert_password = None
        self.timestamp_url = None
        self.sha256 = False
        self._resources = []

    def _payload_name(self) -> str:
        """Get the payload archive name for this package."""
        return f"{self.name}.cab"

    def add_resource(self, name: str, location: str) -> None:
        """Add a public resource provided by this package."""
        self._resources.append((name, location))

    def _create_module(self) -> ElementTree.Element:
        """Create the module document."""
        module = super()._create_module()
        module_intended_client = ElementTree.SubElement(module, "intendedClient")
        module_intended_client.text = "Forms"
        module_intended_platform = ElementTree.SubElement(module, "intendedPlatform")
        module_intended_platform.text = "WINDOWS_X64"
        module_resources = ElementTree.SubElement(module, "resources")
        for resource_name, resource_location in self._resources:
            ElementTree.SubElement(module_resources, "publicResource", {
                "name": resource_name,
                "relativePath": resource_location
            })
        return module

    def _build_payload(self, metadata: ElementTree.Element, module: ElementTree.Element, payload_dest: str) -> None:
        """Build the main payload archive for the SPK package."""
        # pylint: disable=import-outside-toplevel
        from spotfire import cabfile, codesign
        metadata_files = metadata.find("Files")

        with cabfile.CabFile(payload_dest) as payload:
            # Add all files that are supposed to go into the package
            for filename_ondisk, filename_payload in self._contents:
                payload.write(filename_ondisk, filename_payload)
                stat = os.lstat(filename_ondisk)
                ElementTree.SubElement(metadata_files, "File", {
                    "TargetRelativePath": filename_payload,
                    "LastModifiedDate": datetime.datetime.utcfromtimestamp(stat.st_ctime).strftime(
                        "%Y-%m-%dT%H:%M:%SZ"),
                })

            # Add the module.xml file
            payload.writestr("module.xml", _et_to_bytes(module))

        # Codesign the payload
        if self.cert_file:
            codesign.codesign_file(payload_dest, self.cert_file, self.cert_password, self.timestamp_url, self.sha256,
                                   self.sha256)


# Subcommands

@subcommand([argument("spk-file", help="path to the SPK file to build"),
             argument("-v", "--version", default=1000, type=int, help="set the final component of the version number "
                                                                      "of the built SPK file (default: 1000)"),
             argument("--exclude", action="append", metavar="PATH", help="exclude files from the built SPK file"),
             argument("-c", "--constraint", metavar="FILE", help="apply the constraints in the file when installing "
                                                                 "Python packages"),
             argument("--analyst", action="store_true", help="build the SPK file for use with Spotfire Analyst"),
             argument("--cert", metavar="FILE", help="path to the certificate file to sign the package with (Analyst "
                                                     "only)"),
             argument("--password", help="password for the certificate file (Analyst only)"),
             argument("--timestamp", metavar="URL", help="URL of a timestamping service to timestamp the package with "
                                                         "(Analyst only)"),
             argument("--sha256", action="store_true", help="use SHA-256 for file and timestamp digests (Analyst only)")
             ])
def python(args, hook=None) -> None:
    """Package the currently running Python interpreter as an SPK package"""

    # Verify the version component
    version_identifier = getattr(args, "version")
    if version_identifier < 0:
        print("Error: '--version' cannot be less than 0.")
        sys.exit(1)
    if version_identifier < 1000:
        print("Warning: '--version' should not be less than 1000 in order to avoid conflicts with TIBCO-provided "
              "packages.")

    # Set up the package builder
    analyst = getattr(args, "analyst")
    if analyst:
        package_builder = _CabPackageBuilder()
        package_builder.excludes = getattr(args, "exclude")
        package_builder.cert_file = getattr(args, "cert")
        package_builder.cert_password = getattr(args, "password")
        package_builder.timestamp_url = getattr(args, "timestamp")
        package_builder.sha256 = getattr(args, "sha256")
    else:
        package_builder = _ZipPackageBuilder()
        package_builder.excludes = getattr(args, "exclude")
        package_builder.chmod_script_name = "python_chmod"
    package_builder.output = getattr(args, "spk-file")

    # Determine the platform and any constants that depend on it
    if analyst:
        package_builder.id = "2fe5bbd2-9748-40e5-95b5-748b556ed822"
        package_builder.name = "Python Interpreter"
    else:
        package_builder.id = {
            "Windows": "b95fbe51-c013-4f65-8523-5bffcf19e6a8",
            "Linux": "6692a2c3-d43d-4224-a8db-26619ae8f268",
        }.get(platform.system())
        package_builder.name = f"Python Interpreter {platform.system()}"

    # Get the version of the Python installation
    if sys.version_info.major == 3 and sys.version_info.minor < 5:
        print(f"Error: Unsupported version of Python ('{sys.version}').")
        sys.exit(1)
    package_builder.version = _SpkVersion.from_version_info(version_identifier)

    # Scan the files required to create the Python interpreter SPK
    def cleanup():
        """Dummy cleanup function."""

    if analyst:
        prefix = "python"
        package_builder.add_resource("default.python.interpreter.directory", prefix)
    else:
        prefix = "root/python"
    package_builder.scan_python_installation(prefix)
    package_builder.scan_spotfire_package(prefix)
    spotfire_requirements = os.path.join(spotfire.__path__[0], "requirements.txt")
    try:
        constraints = getattr(args, "constraint")
        cleanup, _ = package_builder.scan_requirements_txt(spotfire_requirements, constraints, prefix)
        if hook is not None:
            hook.scan_finished(package_builder)

        # Build the package
        package_builder.build()
    finally:
        cleanup()


@subcommand([argument("spk-file", help="path to the SPK file to build"),
             argument("requirements", help="package the Python packages listed in the file"),
             argument("-v", "--version", help="set the version number of the built SPK file"),
             argument("-f", "--force", action="store_true", help="ignore errors about downgrading version numbers"),
             argument("--versioned-filename", action="store_true", help="modify the filename of the SPK file to "
                                                                        "contain the version number of the built "
                                                                        "package"),
             argument("-n", "--name", help="set the internal module name of the built SPK file"),
             argument("-c", "--constraint", metavar="FILE", help="apply the constraints in the file when installing "
                                                                 "Python packages"),
             argument("--analyst", action="store_true", help="build the SPK file for use with Spotfire Analyst"),
             argument("--cert", metavar="FILE", help="path to the certificate file to sign the package with (Analyst "
                                                     "only)"),
             argument("--password", help="password for the certificate file (Analyst only)"),
             argument("--timestamp", metavar="URL", help="URL of a timestamping service to timestamp the package with "
                                                         "(Analyst only)"),
             argument("--sha256", action="store_true", help="use SHA-256 for file and timestamp digests (Analyst only)")
             ])
def packages(args) -> None:
    """Package a list of Python packages as an SPK package"""

    def cleanup():
        """Dummy cleanup function."""

    try:
        # Set up the package builder
        analyst = getattr(args, "analyst")
        if analyst:
            package_builder = _CabPackageBuilder()
            package_builder.cert_file = getattr(args, "cert")
            package_builder.cert_password = getattr(args, "password")
            package_builder.timestamp_url = getattr(args, "timestamp")
            package_builder.sha256 = getattr(args, "sha256")
        else:
            package_builder = _ZipPackageBuilder()
            package_builder.chmod_script_name = "packages_chmod"
        package_builder.output = getattr(args, "spk-file")
        requirements_file = getattr(args, "requirements")
        brand = _read_brand(requirements_file, "## spotfire.spk: ")
        version = getattr(args, "version")
        force = getattr(args, "force")
        versioned_filename = getattr(args, "versioned_filename")
        name = getattr(args, "name") or brand.get("BuiltName")
        pkg_id = brand.get("BuiltId")

        # If name and id are not in the brand or given on the command line, generate reasonable defaults
        if name is None:
            if analyst:
                name = "Python Packages"
            else:
                name = f"Python Packages {platform.system()}"
        if pkg_id is None:
            pkg_id = str(uuid.uuid4())
        package_builder.name = name
        package_builder.id = pkg_id

        # Scan the requirements file for packages to install
        if analyst:
            prefix = "site-packages"
            prefix_direct = True
            package_builder.add_resource(f"python.package.{package_builder.id}.whl", prefix)
        else:
            prefix = "root/python"
            prefix_direct = False
        constraints = getattr(args, "constraint")
        cleanup, installed_packages = package_builder.scan_requirements_txt(requirements_file, constraints, prefix,
                                                                            prefix_direct, True)

        # Based on the packages we installed, determine how to increment the version number
        _handle_versioning(package_builder, installed_packages, brand, version, force, versioned_filename)

        # Build the package
        package_builder.build()

        # Now prepare the brand with the results of the build and apply it to our requirements file
        brand["BuiltBy"] = sys.version
        brand["BuiltAt"] = time.asctime(time.localtime(os.path.getmtime(package_builder.output)))
        brand["BuiltFile"] = package_builder.output
        brand["BuiltName"] = package_builder.name
        brand["BuiltId"] = package_builder.id
        brand["BuiltVersion"] = str(package_builder.version)
        brand["BuiltPackages"] = installed_packages
        _brand_file(requirements_file, brand, "## spotfire.spk: ")
    finally:
        cleanup()


def _handle_versioning(package_builder, installed_packages, brand, version, force, versioned_filename):
    package_builder.version = _SpkVersion()
    if "BuiltVersion" in brand:
        package_builder.version = _SpkVersion.from_str(brand["BuiltVersion"])
        package_builder.version.increment_minor()
        if "BuiltPackages" in brand:
            # Tick the major version if required
            if _should_increment_major(brand["BuiltPackages"], installed_packages, force):
                package_builder.version.increment_major()
    # Handle manually specified version numbers
    if version:
        given_version = _SpkVersion.from_str(version)
        if given_version < package_builder.version:
            _error(f"Package version '{given_version}' is lower than generated version '{package_builder.version}'.")
            if not force:
                sys.exit(1)
        package_builder.version = given_version
    # Handle versioned filenames
    if versioned_filename:
        package_builder.output = re.sub(r"(\.spk)?$", fr"-{str(package_builder.version)}\1",
                                        package_builder.output, 1)


def _should_increment_major(old_packages, new_packages, force):
    tick_major = False
    # Check for removed packages
    if old_packages.keys() - new_packages.keys():
        tick_major = True
    # Check for package downgrades
    for pkg in old_packages.keys() & new_packages.keys():
        previous_version = pip_version.parse(old_packages[pkg])
        new_version = pip_version.parse(new_packages[pkg])
        if previous_version > new_version:
            tick_major = True
            _error(f"Package '{pkg}' has a lower version than previously built.")
            if not force:
                sys.exit(1)
    return tick_major


# Main

def main() -> None:
    """Start the package builder."""
    cli_args = CLI_PARSER.parse_args()
    if cli_args.subcommand is None:
        CLI_PARSER.print_help()
    else:
        cli_args.func(cli_args)


if __name__ == '__main__':
    main()
    sys.exit(0)



# Copyright © 2021. TIBCO Software Inc.
# This file is subject to the license terms contained
# in the license file that is distributed with this file.

"""Utilities used by multiple submodules."""
import os
import tempfile


def type_name(type_: type) -> str:
    """Convert a type object to a string in a consistent manner.

    :param type_: the type object to convert
    :return: a string with the type name
    """
    type_qualname = type_.__qualname__
    type_module = type_.__module__
    if type_module not in ("__main__", "builtins"):
        type_qualname = type_module + '.' + type_qualname
    return type_qualname


class TempFiles:
    """Utility class that manages the lifecycle of multiple temporary files."""
    def __init__(self) -> None:
        self._files = []

    def new_file(self, **kwargs) -> tempfile.NamedTemporaryFile:
        """Create a temporary file object that will be tracked by this manager.

        :param kwargs: NamedTemporaryFile arguments
        :return: a new NamedTemporaryFile object
        """
        # pylint: disable=consider-using-with
        temp_file = tempfile.NamedTemporaryFile(delete=False, **kwargs)
        self._files.append(temp_file)
        return temp_file

    def cleanup(self) -> None:
        """Clean up all managed temporary files by deleting them."""
        for temp_file in self._files:
            os.unlink(temp_file.name)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.cleanup()



# Copyright © 2021. TIBCO Software Inc.
# This file is subject to the license terms contained
# in the license file that is distributed with this file.

"""Tools to apply Authenticode code signing signatures and timestamps to files using the native Microsoft APIs.
Only defined on Windows platforms."""

import sys

if sys.platform == "win32":
    # pylint: disable=invalid-name,too-few-public-methods,attribute-defined-outside-init
    import ctypes
    from ctypes import wintypes
    import os
    import typing

    # Constant declarations
    _S_OK = 0

    _X509_ASN_ENCODING = 0x00010000
    _PKCS_7_ASN_ENCODING = 0x00000001

    _CERT_FIND_ANY = 0
    _CERT_KEY_SPEC_PROP_ID = 6
    _CERT_CLOSE_STORE_CHECK_FLAG = 2

    _SIGNER_SUBJECT_FILE = 1
    _SIGNER_CERT_POLICY_CHAIN = 2
    _SIGNER_CERT_STORE = 2
    _SIGNER_AUTHCODE_ATTR = 1

    _SIGNER_TIMESTAMP_AUTHENTICODE = 1
    _SIGNER_TIMESTAMP_RFC3161 = 2

    _CALG_SHA_256 = 0x800C
    _CALG_SHA1 = 0x8004
    _OID_SHA256 = b"2.16.840.1.101.3.4.2.1"
    _OID_SHA1 = b"1.3.14.3.2.26"

    # Type declarations
    _PCCERT_CONTEXT = ctypes.c_void_p
    _PSIGNER_CONTEXT = ctypes.c_void_p
    _PSIGNER_PROVIDER_INFO = ctypes.c_void_p
    _ALG_ID = ctypes.c_uint
    _PCRYPT_ATTRIBUTES = ctypes.c_void_p

    class _CRYPT_DATA_BLOB(ctypes.Structure):
        _fields_ = [("cbData", wintypes.DWORD),
                    ("pbData", ctypes.c_char_p)]

    class _SIGNER_FILE_INFO(ctypes.Structure):
        _fields_ = [("cbSize", wintypes.DWORD),
                    ("pwszFileName", wintypes.LPCWSTR),
                    ("hFile", wintypes.HANDLE)]

    class _SIGNER_SUBJECT_INFO(ctypes.Structure):
        _fields_ = [("cbSize", wintypes.DWORD),
                    ("pdwIndex", wintypes.LPDWORD),
                    ("dwSubjectChoice", wintypes.DWORD),
                    ("pSignerFileInfo", ctypes.POINTER(_SIGNER_FILE_INFO))]

    class _SIGNER_CERT_STORE_INFO(ctypes.Structure):
        _fields_ = [("cbSize", wintypes.DWORD),
                    ("pSigningCert", _PCCERT_CONTEXT),
                    ("dwCertPolicy", wintypes.DWORD),
                    ("hCertStore", wintypes.HANDLE)]

    class _SIGNER_CERT(ctypes.Structure):
        _fields_ = [("cbSize", wintypes.DWORD),
                    ("dwCertChoice", wintypes.DWORD),
                    ("pCertStoreInfo", ctypes.POINTER(_SIGNER_CERT_STORE_INFO)),
                    ("hwnd", wintypes.HWND)]

    class _SIGNER_ATTR_AUTHCODE(ctypes.Structure):
        _fields_ = [("cbSize", wintypes.DWORD),
                    ("fCommercial", wintypes.BOOL),
                    ("fIndividual", wintypes.BOOL),
                    ("pwszName", wintypes.LPCWSTR),
                    ("pwszInfo", wintypes.LPCWSTR)]

    class _SIGNER_SIGNATURE_INFO(ctypes.Structure):
        _fields_ = [("cbSize", wintypes.DWORD),
                    ("algidHash", _ALG_ID),
                    ("dwAttrChoice", wintypes.DWORD),
                    ("pAttrAuthcode", ctypes.POINTER(_SIGNER_ATTR_AUTHCODE)),
                    ("psAuthenticated", _PCRYPT_ATTRIBUTES),
                    ("psUnauthenticated", _PCRYPT_ATTRIBUTES)]

    # Parameter declarations
    def _wrap_function(lib, function_name, restype, argtypes):
        """Simplify wrapping ctypes functions"""
        func = lib.__getattr__(function_name)
        func.restype = restype
        func.argtypes = argtypes
        return func


    _pfx_import_cert_store = _wrap_function(ctypes.windll.crypt32,
                                            "PFXImportCertStore",
                                            wintypes.HANDLE,
                                            [ctypes.POINTER(_CRYPT_DATA_BLOB),
                                             wintypes.LPCWSTR,
                                             wintypes.DWORD])

    _cert_close_store = _wrap_function(ctypes.windll.crypt32,
                                       "CertCloseStore",
                                       wintypes.BOOL,
                                       [wintypes.HANDLE,
                                        wintypes.DWORD])

    _cert_find_cert_in_store = _wrap_function(ctypes.windll.crypt32,
                                              "CertFindCertificateInStore",
                                              _PCCERT_CONTEXT,
                                              [wintypes.HANDLE,
                                               wintypes.DWORD,
                                               wintypes.DWORD,
                                               wintypes.DWORD,
                                               ctypes.c_void_p,
                                               _PCCERT_CONTEXT])

    _cert_get_context_property = _wrap_function(ctypes.windll.crypt32,
                                                "CertGetCertificateContextProperty",
                                                wintypes.BOOL,
                                                [_PCCERT_CONTEXT,
                                                 wintypes.DWORD,
                                                 ctypes.c_void_p,
                                                 wintypes.LPDWORD])

    _cert_free_context = _wrap_function(ctypes.windll.crypt32,
                                        "CertFreeCertificateContext",
                                        wintypes.BOOL,
                                        [_PCCERT_CONTEXT])

    _signer_sign_ex = _wrap_function(ctypes.windll.mssign32,
                                     "SignerSignEx",
                                     ctypes.HRESULT,
                                     [wintypes.DWORD,
                                      ctypes.POINTER(_SIGNER_SUBJECT_INFO),
                                      ctypes.POINTER(_SIGNER_CERT),
                                      ctypes.POINTER(_SIGNER_SIGNATURE_INFO),
                                      _PSIGNER_PROVIDER_INFO,
                                      wintypes.LPCWSTR,
                                      _PCRYPT_ATTRIBUTES,
                                      ctypes.c_void_p,
                                      ctypes.POINTER(_PSIGNER_CONTEXT)])

    _signer_timestamp = _wrap_function(ctypes.windll.mssign32,
                                       "SignerTimeStamp",
                                       ctypes.HRESULT,
                                       [ctypes.POINTER(_SIGNER_SUBJECT_INFO),
                                        wintypes.LPCWSTR,
                                        _PCRYPT_ATTRIBUTES,
                                        ctypes.c_void_p])

    _signer_timestamp_ex2 = _wrap_function(ctypes.windll.mssign32,
                                           "SignerTimeStampEx2",
                                           ctypes.HRESULT,
                                           [wintypes.DWORD,
                                            ctypes.POINTER(_SIGNER_SUBJECT_INFO),
                                            wintypes.LPCWSTR,
                                            ctypes.c_char_p,
                                            _PCRYPT_ATTRIBUTES,
                                            ctypes.c_void_p,
                                            ctypes.POINTER(_PSIGNER_CONTEXT)])

    _signer_free_context = _wrap_function(ctypes.windll.mssign32,
                                          "SignerFreeSignerContext",
                                          ctypes.HRESULT,
                                          [_PSIGNER_CONTEXT])

    # Public functions

    # noinspection PyUnboundLocalVariable
    def codesign_file(filename: str,
                      certificate: str,
                      password: typing.Optional[str],
                      timestamp: typing.Optional[str] = None,
                      use_rfc3161: bool = False,
                      use_sha256: bool = False) -> None:
        """Codesign a file using the Microsoft signing API found in mssign32.dll.

        :param filename: the filename of the file to codesign
        :param certificate: the filename of the certificate file to codesign with
        :param password: the password used to unlock the certificate
        :param timestamp: a URL of the timestamping service to timestamp the code signature with
        :param use_rfc3161: whether or not to use the RFC 3161 timestamping protocol.  If ``True``, use RFC 3161.
          If ``False``, use Authenticode.
        :param use_sha256: whether or not to use SHA-256 as the timestamping hash function.  If ``True``, use SHA-256.
          If ``False``, use SHA-1.
        """
        try:
            # Sanity check arguments
            if not os.path.isfile(filename):
                raise FileNotFoundError(f"No such file: '{filename}'")
            if not os.path.isfile(certificate):
                raise FileNotFoundError(f"No such file: '{certificate}'")
            if use_sha256 and not use_rfc3161:
                raise Exception("SHA-256 timestamping requires the RFC 3161 timestamping protocol")

            # Open the certificate file and convert it into an in-memory cert store
            cert_store = _open_certificate(certificate, password)

            # Extract the cert from the new cert store
            cert_context = _extract_certificate(cert_store)

            # Prepare structures
            index = wintypes.DWORD()
            signer_subject_info = _create_subject(index, filename)

            # Sign the file
            _sign(cert_context, use_sha256, signer_subject_info)

            # Timestamp the file
            _timestamp(signer_subject_info, timestamp, use_rfc3161, use_sha256)

        finally:
            try:
                if cert_context is not None:
                    _cert_free_context(cert_context)
            except NameError:
                pass

            try:
                if cert_store is not None:
                    _cert_close_store(cert_store, _CERT_CLOSE_STORE_CHECK_FLAG)
            except NameError:
                pass


    def _open_certificate(certificate, password):
        """Open the certificate file and convert it into an in-memory cert store."""
        with open(certificate, "rb") as cert:
            cert_data = cert.read()
        cert_blob = _CRYPT_DATA_BLOB(len(cert_data), cert_data)
        cert_store = _pfx_import_cert_store(ctypes.byref(cert_blob), password, 0)
        if cert_store is None:
            cert_store = _pfx_import_cert_store(ctypes.byref(cert_blob), "", 0)
        if cert_store is None:
            cert_store = _pfx_import_cert_store(ctypes.byref(cert_blob), None, 0)
        if cert_store is None:
            raise Exception(f"Could not load certificate; is the password correct? (0x{ctypes.GetLastError():08x})")
        return cert_store


    def _extract_certificate(cert_store):
        """Extract the cert from the new cert store."""
        cert_context = _cert_find_cert_in_store(cert_store, _X509_ASN_ENCODING | _PKCS_7_ASN_ENCODING, 0,
                                                _CERT_FIND_ANY, None, None)
        if cert_context is None:
            raise Exception(f"Could not get certificate from store (0x{ctypes.GetLastError():08x})")
        key_spec = wintypes.DWORD()
        found_private_key = False
        while not found_private_key:
            key_spec_len = wintypes.DWORD(ctypes.sizeof(key_spec))
            has_private_key = _cert_get_context_property(cert_context, _CERT_KEY_SPEC_PROP_ID,
                                                         ctypes.byref(key_spec), ctypes.byref(key_spec_len))
            if has_private_key:
                found_private_key = True
            else:
                cert_context = _cert_find_cert_in_store(cert_store, _X509_ASN_ENCODING | _PKCS_7_ASN_ENCODING, 0,
                                                        _CERT_FIND_ANY, None, cert_context)
                if cert_context is None:
                    raise Exception(f"Could not get certificate from store (0x{ctypes.GetLastError():08x})")
        return cert_context


    def _create_subject(index, filename):
        """Create a subject information structure for a given file."""
        signer_file_info = _SIGNER_FILE_INFO()
        signer_file_info.cbSize = ctypes.sizeof(_SIGNER_FILE_INFO)
        signer_file_info.pwszFileName = filename
        signer_file_info.hFile = None

        signer_subject_info = _SIGNER_SUBJECT_INFO()
        signer_subject_info.cbSize = ctypes.sizeof(_SIGNER_SUBJECT_INFO)
        signer_subject_info.pdwIndex = ctypes.pointer(index)
        signer_subject_info.dwSubjectChoice = _SIGNER_SUBJECT_FILE
        signer_subject_info.pSignerFileInfo = ctypes.pointer(signer_file_info)

        return signer_subject_info


    def _sign(cert_context, use_sha256, signer_subject_info):
        """Sign the file."""
        signer_cert_store_info = _SIGNER_CERT_STORE_INFO()
        signer_cert_store_info.cbSize = ctypes.sizeof(_SIGNER_CERT_STORE_INFO)
        signer_cert_store_info.pSigningCert = cert_context
        signer_cert_store_info.dwCertPolicy = _SIGNER_CERT_POLICY_CHAIN
        signer_cert_store_info.hCertStore = None

        signer_cert = _SIGNER_CERT()
        signer_cert.cbSize = ctypes.sizeof(_SIGNER_CERT)
        signer_cert.dwCertChoice = _SIGNER_CERT_STORE
        signer_cert.pCertStoreInfo = ctypes.pointer(signer_cert_store_info)
        signer_cert.hwnd = None

        signer_attr_authcode = _SIGNER_ATTR_AUTHCODE()
        signer_attr_authcode.cbSize = ctypes.sizeof(_SIGNER_ATTR_AUTHCODE)
        signer_attr_authcode.fCommercial = False
        signer_attr_authcode.fIndividual = True
        signer_attr_authcode.pwszName = ""
        signer_attr_authcode.pwszInfo = None

        signer_signature_info = _SIGNER_SIGNATURE_INFO()
        signer_signature_info.cbSize = ctypes.sizeof(_SIGNER_SIGNATURE_INFO)
        signer_signature_info.algidHash = _CALG_SHA_256 if use_sha256 else _CALG_SHA1
        signer_signature_info.dwAttrChoice = _SIGNER_AUTHCODE_ATTR
        signer_signature_info.pAttrAuthcode = ctypes.pointer(signer_attr_authcode)
        signer_signature_info.psAuthenticated = None
        signer_signature_info.psUnauthenticated = None

        signer_context = _PSIGNER_CONTEXT()
        result = _signer_sign_ex(0,
                                 ctypes.byref(signer_subject_info),
                                 ctypes.byref(signer_cert),
                                 ctypes.byref(signer_signature_info),
                                 None,
                                 None,
                                 None,
                                 None,
                                 ctypes.byref(signer_context))
        if signer_context is not None:
            _signer_free_context(signer_context)
        if result is not _S_OK:
            raise Exception(f"Could not sign file (0x{ctypes.GetLastError():08x})")


    def _timestamp(signer_subject_info, timestamp, use_rfc3161, use_sha256):
        """Timestamp the file."""
        if timestamp is not None:
            if use_rfc3161:
                signer_context = _PSIGNER_CONTEXT()
                result = _signer_timestamp_ex2(_SIGNER_TIMESTAMP_RFC3161,
                                               ctypes.byref(signer_subject_info),
                                               timestamp,
                                               _OID_SHA256 if use_sha256 else _OID_SHA1,
                                               None,
                                               None,
                                               ctypes.byref(signer_context))
                if signer_context is not None:
                    _signer_free_context(signer_context)
            else:
                result = _signer_timestamp(ctypes.byref(signer_subject_info), timestamp, None, None)
            if result is not _S_OK:
                raise Exception(f"Could not timestamp file (0x{ctypes.GetLastError():08x})")



# Copyright © 2021. TIBCO Software Inc.
# This file is subject to the license terms contained
# in the license file that is distributed with this file.

"""User visible utility functions."""

import warnings

import pandas as pd


# Table and column metadata functions

def copy_metadata(source, destination) -> None:
    """Copy the table and column metadata from a Pandas object to another.
    :param source: the DataFrame or Series to copy metadata from
    :param destination: the DataFrame or Series to copy metadata to
    :raise TypeError: if the types of source and destination do not match
    """
    # Verify that types of source and destination match
    if isinstance(source, pd.DataFrame) and not isinstance(destination, pd.DataFrame):
        raise TypeError("both source and destination must be DataFrames")
    if isinstance(source, pd.Series) and not isinstance(destination, pd.Series):
        raise TypeError("both source and destination must be Series")

    # Handle DataFrames
    if isinstance(source, pd.DataFrame):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            destination.spotfire_table_metadata = source.spotfire_table_metadata
        for col in source.columns:
            try:
                source1 = source[col]
                destination1 = destination[col]
                destination1.spotfire_column_metadata = source1.spotfire_column_metadata
            except AttributeError:
                pass
    # Handle Series
    elif isinstance(source, pd.Series):
        destination.spotfire_column_metadata = source.spotfire_column_metadata



# Copyright © 2021. TIBCO Software Inc.
# This file is subject to the license terms contained
# in the license file that is distributed with this file.

"""Tools to create Microsoft cabinet files.  Only defined on Windows platforms."""

import sys

if sys.platform == "win32":
    import msilib
    import os
    import typing
    from spotfire import _utils

    class CabFile:
        """Class with methods to open, write, and close Microsoft cabinet files."""
        def __init__(self, file: str) -> None:
            """Open the cabinet file.

            :param file: filename of the cabinet file
            """
            self.filename = file
            self._contents = []
            self._temp_files = _utils.TempFiles()
            self._opened = True

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            return self.close()

        def __repr__(self) -> str:
            result = [f"<{self.__class__.__module__}.{self.__class__.__qualname__}"]
            if not self._opened:
                result.append(" [closed]")
            elif self.filename is not None:
                result.append(f" filename={self.filename!r}")
            result.append(">")
            return "".join(result)

        def write(self, filename: str, arcname=None) -> None:
            """Put a file into the cabinet.

            :param filename: the filename of the file to put in the cabinet
            :param arcname: the name in the cabinet the file should be placed under.  If ``None``, use the basename
             (without path information) of filename.
            """
            if not self._opened:
                raise ValueError("Attempt to write to cabinet that was already closed")

            if arcname is None:
                arcname = os.path.basename(filename)
            self._contents.append((filename, arcname))

        def writestr(self, arcname: str, data: typing.Union[str, bytes]) -> None:
            """Write arbitrary data into the cabinet.

            :param arcname: the name in the cabinet the data should be placed under
            :param data: the data to place in the cabinet
            """
            if not self._opened:
                raise ValueError("Attempt to write to cabinet that was already closed")

            # write the data out to a temp file
            temp = self._temp_files.new_file()
            temp.file.write(data)
            temp.close()
            self._contents.append((temp.name, arcname))

        def close(self) -> None:
            """Close the cabinet and write the contents to disk."""
            if self._opened:
                msilib.FCICreate(self.filename, self._contents)
                self._temp_files.cleanup()
                self._opened = False



# Copyright © 2021. TIBCO Software Inc.
# This file is subject to the license terms contained
# in the license file that is distributed with this file.

# pylint: skip-file
from setuptools import setup, find_packages


def get_requires(filename):
    requirements = []
    with open(filename, "rt") as req_file:
        for line in req_file.read().splitlines():
            if not line.strip().startswith("#"):
                requirements.append(line)
    return requirements


project_requirements = get_requires("spotfire/requirements.txt")
version = {}
with open('spotfire/version.py') as ver_file:
    exec(ver_file.read(), version)
    
with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='spotfire',
    version=version['__version__'],
    description='Package for Building Python Extensions to TIBCO Spotfire',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='TIBCO Software Inc.',
    maintainer='Spotfire Python Package Support',
    maintainer_email='spotfirepython@tibco.com',
    url='https://github.com/TIBCOSoftware/spotfire-python',
    license='BSD 3-Clause License',
    packages=find_packages(exclude=['spotfire.test']),
    include_package_data=True,
    install_requires=project_requirements,
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
)



# Copyright © 2021. TIBCO Software Inc.
# This file is subject to the license terms contained
# in the license file that is distributed with this file.

"""Functions to import data tables into Python from SBDF files and export data from Python to SBDF files."""

import collections.abc
import datetime
import decimal
import enum
import struct
import tempfile
import typing
import warnings

import bitstring
import pandas as pd
import numpy as np

from spotfire import _utils

try:
    import geopandas as gpd
    import shapely
    import shapely.geometry.base as shp_geom
except ImportError:
    gpd = None
    shapely = None
    shp_geom = None

try:
    import matplotlib
    import matplotlib.figure
except ImportError:
    matplotlib = None

try:
    import PIL
    import PIL.Image
except ImportError:
    PIL = None

try:
    import seaborn
except ImportError:
    seaborn = None

__all__ = ["import_data", "export_data"]


# Public Functions


def import_data(sbdf_file: typing.Union[str, bytes, int]) -> pd.DataFrame:
    """Import data from an SBDF file and create a 'pandas' DataFrame.

    :param sbdf_file: the filename of the SBDF file to import
    :return: the DataFrame containing the imported data
    :raises SBDFError: if a problem is encountered during import
    """

    # Open the SBDF file
    with open(sbdf_file, "rb") as file:
        # Read the file header
        version_major, version_minor = _FileHeader.read(file)
        if version_major != _FileHeader.Major_Version or version_minor != _FileHeader.Minor_Version:
            raise SBDFError(f"unsupported file version {version_major}.{version_minor}")

        # Read the table metadata
        tmeta = _TableMetadata.read(file)

        # Process table metadata
        table_metadata_dict = _import_table_metadata(tmeta)

        # Process column metadata
        pd_data, pd_dtypes, column_metadata_dict, column_names = _import_column_metadata(tmeta)

        # Read the table slices
        _import_table_slices(file, column_names, pd_data, tmeta)

        # Construct the pandas DataFrame and return the final results
        columns = []
        for col in column_names:
            columns.append(pd.Series(pd_data[col], dtype=pd_dtypes[col], name=col))
        dataframe = pd.concat(columns, axis=1)
        for col in column_names:
            dataframe[col].spotfire_column_metadata = column_metadata_dict[col]
        if gpd is not None and table_metadata_dict.get('MapChart.IsGeocodingTable'):
            dataframe = _data_frame_to_geo_data_frame(dataframe, table_metadata_dict)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            dataframe.spotfire_table_metadata = table_metadata_dict
        return dataframe


def _import_table_metadata(tmeta: '_TableMetadata') -> typing.Dict[str, typing.Any]:
    table_metadata_dict = {}
    for i in range(tmeta.table_metadata.count()):
        table_metadata_dict[tmeta.table_metadata.names[i]] = tmeta.table_metadata.values[i].data
    return table_metadata_dict


def _import_column_metadata(tmeta: '_TableMetadata') -> typing.Tuple[typing.Dict[str, typing.List],
                                                                     typing.Dict[str, str],
                                                                     typing.Dict[str, typing.Dict[str, typing.Any]],
                                                                     typing.List[str]]:
    pd_data = {}
    pd_dtypes = {}
    column_metadata_dict = {}
    column_names = []
    for i in range(tmeta.column_count()):
        cmeta = tmeta.column_metadata[i]

        # get the column name
        cm_column_name = _ColumnMetadata.get_name(cmeta)
        column_names.append(cm_column_name)

        # add a new list to hold the column
        pd_data[cm_column_name] = []

        # get the pandas dtype for constructing this column
        pd_dtypes[cm_column_name] = _ColumnMetadata.get_type(cmeta).to_dtype_name()

        # get the remaining column metadata
        cm_dict = {}
        for j in range(cmeta.count()):
            if cmeta.names[j] in (_ColumnMetadata.Property_Name, _ColumnMetadata.Property_DataType):
                continue
            cm_dict[cmeta.names[j]] = cmeta.values[j].data
        column_metadata_dict[cm_column_name] = cm_dict

    return pd_data, pd_dtypes, column_metadata_dict, column_names


def _import_table_slices(file: typing.BinaryIO, column_names: typing.List[str],
                         pd_data: typing.Dict[str, typing.List], tmeta: '_TableMetadata') -> None:
    while True:
        tslice = _TableSlice.read(file, tmeta, None)
        if tslice is None:
            break
        for i in range(tslice.column_count()):
            cslice = tslice.columns[i]
            cs_values = cslice.values.get_values()
            cs_invalid_prop = cslice.get_property(_ColumnSlice.ValueProperty_IsInvalid)
            if cs_invalid_prop is None:
                cs_invalid = [False] * cs_values.get_count()
            else:
                cs_invalid = cs_invalid_prop.get_values().data
            for value, invalid in zip(cs_values.data, cs_invalid):
                pd_data[column_names[i]].append(None if invalid else value)


def export_data(obj: typing.Any, sbdf_file: typing.Union[str, bytes, int], default_column_name: str = "x") -> None:
    """Export data to an SBDF file.

    :param obj: the data object to export
    :param sbdf_file: the filename to export the data to
    :param default_column_name: the column name to use when exporting data objects without intrinsic names (such as
                                lists or scalar values)
    :raises SBDFError: if a problem is encountered during export
    """
    columns, column_names, column_types, table_metadata, column_metadata = _export_columnize_data(obj,
                                                                                                  default_column_name)

    # Open the SBDF file
    with open(sbdf_file, "wb") as file:
        # Write the file header
        _FileHeader.write(file)

        # Write the table and column metadata
        tmeta = _export_table_metadata(table_metadata)
        row_count = _export_column_metadata(columns, column_names, column_types, column_metadata, tmeta)
        tmeta.write(file)

        # Write out the table and column slices
        _export_table_slices(columns, column_names, column_types, file, row_count, tmeta)


def _export_columnize_data(obj: typing.Any, default_column_name: str) -> \
        typing.Tuple[typing.Dict[str, typing.List[typing.Any]],
                     typing.List[str],
                     typing.Dict[str, '_ValueTypeId'],
                     typing.Dict[str, typing.List[typing.Any]],
                     typing.Dict[str, typing.Dict[str, typing.List[typing.Any]]]
                     ]:
    # pylint: disable=too-many-branches,too-many-statements
    table_metadata = {}
    column_metadata = {}
    if isinstance(obj, pd.DataFrame):
        # Extract the table and column metadata from the data frame
        try:
            table_metadata = obj.spotfire_table_metadata
        except AttributeError:
            pass
        for col in obj.columns.tolist():
            try:
                col_meta = obj[col].spotfire_column_metadata
                column_metadata[col] = col_meta
            except AttributeError:
                column_metadata[col] = {}

        # Convert geopandas geodataframe to Spotfire's native geocoding format
        if gpd is not None and isinstance(obj, gpd.GeoDataFrame):
            obj = _geo_data_frame_to_data_frame(obj, table_metadata, column_metadata)

        if len({str(x) for x in obj.keys()}) != len(obj.columns):
            raise SBDFError("obj does not have unique column names")
        columns = obj.to_dict("list")
        column_names = obj.columns.tolist()
        column_types = {str(k): _ValueTypeId.infer_from_dtype(v, f"column '{str(k)}'") for (k, v) in obj.iteritems()}
    elif isinstance(obj, pd.Series):
        # Handle series as columnar data
        series_name = default_column_name if obj.name is None else obj.name
        series_description = "series" if obj.name is None else f"series '{obj.name}'"

        # Extract the column metadata from the series
        try:
            column_metadata = {series_name: obj.spotfire_column_metadata}
        except AttributeError:
            pass

        columns = {series_name: obj.tolist()}
        column_names = [series_name]
        column_types = {series_name: _ValueTypeId.infer_from_dtype(obj, series_description)}
    elif isinstance(obj, dict):
        # Handle dicts of lists as tabular data
        if not all(isinstance(x, list) for x in obj.values()):
            raise SBDFError("obj is not a dict of lists")
        if len({str(x) for x in obj.keys()}) != len(obj):
            raise SBDFError("obj does not have unique column names")
        column_metadata = {col: {} for col in obj.keys()}
        columns = obj
        column_names = obj.keys()
        column_types = {str(k): _ValueTypeId.infer_from_type(v, f"column '{str(k)}'") for (k, v) in obj.items()}
    elif isinstance(obj, (str, bytes, bytearray)):
        # Handle strings and bytes as scalar data
        column_metadata[default_column_name] = {}
        columns = {default_column_name: [obj]}
        column_names = [default_column_name]
        column_types = {default_column_name: _ValueTypeId.infer_from_type([obj], "value")}
    elif isinstance(obj, collections.abc.Iterable):
        # Handle all iterable items as columnar data
        column_metadata[default_column_name] = {}
        columns = {default_column_name: list(obj)}
        column_names = [default_column_name]
        column_types = {default_column_name: _ValueTypeId.infer_from_type(list(obj), "list")}
    else:
        # Handle various image formats
        if matplotlib is not None and isinstance(obj, matplotlib.figure.Figure):
            obj = _pyplot_to_binary(obj)
        elif seaborn is not None and isinstance(obj, seaborn.axisgrid.Grid):
            obj = _seaborn_plot_to_binary(obj)
        elif PIL is not None and isinstance(obj, PIL.Image.Image):
            obj = _pil_image_to_binary(obj)

        # If all else fails, treat this as scalar data
        column_metadata[default_column_name] = {}
        columns = {default_column_name: [obj]}
        column_names = [default_column_name]
        column_types = {default_column_name: _ValueTypeId.infer_from_type([obj], "value")}

    return columns, column_names, column_types, table_metadata, column_metadata


def _export_table_metadata(table_metadata: typing.Dict[str, typing.Any]) -> '_TableMetadata':
    metadata = _Metadata()
    if isinstance(table_metadata, dict):
        for key, val in table_metadata.items():
            obj_val = val if isinstance(val, list) else [val]
            obj_valuetype = _ValueTypeId.infer_from_type(obj_val, "table metadata")
            obj = _SbdfObject(obj_valuetype, obj_val)
            metadata.add(key, obj)
    tmeta = _TableMetadata(metadata)
    return tmeta


def _export_column_metadata(columns: typing.Dict[str, typing.List], column_names: typing.List[str],
                            column_types: typing.Dict[str, '_ValueTypeId'],
                            column_metadata: typing.Dict[str, typing.Dict[str, typing.Any]],
                            tmeta: '_TableMetadata') -> int:
    row_count = None
    for colname in column_names:
        metadata = _Metadata()
        if isinstance(column_metadata, dict):
            colmeta = column_metadata.get(colname, {})
            if not isinstance(colmeta, dict):
                raise SBDFError("column_metadata is not a dict of dicts")
            for key, val in colmeta.items():
                obj_val = val if isinstance(val, list) else [val]
                obj_valuetype = _ValueTypeId.infer_from_type(obj_val, "column metadata")
                obj = _SbdfObject(obj_valuetype, obj_val)
                metadata.add(key, obj)
        if row_count is None:
            row_count = len(columns[colname])
        else:
            if row_count != len(columns[colname]):
                raise SBDFError(f"column '{colname}' has inconsistent column length")
        _ColumnMetadata.set_values(metadata, colname, column_types.get(str(colname)))
        tmeta.add(metadata)
    return row_count


def _export_table_slices(columns: typing.Dict[str, typing.List], column_names: typing.List[str],
                         column_types: typing.Dict[str, '_ValueTypeId'], file: typing.BinaryIO,
                         row_count: int, tmeta: '_TableMetadata') -> None:
    max_rows_per_slice = max(10, 100000 // max(1, len(column_names)))
    row_offset = 0
    while row_offset < row_count:
        slice_row_count = min(max_rows_per_slice, row_count - row_offset)
        tslice = _TableSlice(tmeta)
        for colname in column_names:
            obj = _SbdfObject(column_types.get(str(colname)), columns[colname][row_offset:row_offset + slice_row_count])
            cslice = _ColumnSlice(_ValueArray(_ValueArrayEncoding.PLAIN_ARRAY, obj))
            invalid = [pd.isnull(x) for x in obj.data]
            if any(invalid):
                obj_vt = _ValueType(column_types.get(str(colname)))
                obj.data = [obj_vt.missing_value() if missing else val for val, missing in zip(obj.data, invalid)]
                obj_empty = _SbdfObject(_ValueTypeId.BOOL, invalid)
                va_empty = _ValueArray(_ValueArrayEncoding.BIT_ARRAY, obj_empty)
                cslice.add_property(_ColumnSlice.ValueProperty_IsInvalid, va_empty)
            tslice.add(cslice)
        tslice.write(file)
        row_offset += slice_row_count
    _TableSlice.write_end(file)


# Exceptions

class SBDFError(Exception):
    """An exception that is raised to indicate a problem during import or export of SBDF files."""


# File Headers

class _FileHeader:
    Major_Version = 1
    Minor_Version = 0
    Version_String = "1.0"

    @staticmethod
    def write(file: typing.BinaryIO) -> None:
        """writes the current sbdf fileheader to file"""
        _section_write(file, _SectionTypeId.FILEHEADER)
        _write_int8(file, _FileHeader.Major_Version)
        _write_int8(file, _FileHeader.Minor_Version)

    @staticmethod
    def read(file: typing.BinaryIO) -> typing.Tuple[int, int]:
        """reads the sbdf fileheader from file"""
        _section_expect(file, _SectionTypeId.FILEHEADER)
        major = _read_int8(file)
        minor = _read_int8(file)
        return major, minor


# Column Slices

class _ColumnSlice:
    ValueProperty_IsInvalid = "IsInvalid"
    ValueProperty_ErrorCode = "ErrorCode"
    ValueProperty_ReplacedValue = "HasReplacedValue"

    def __init__(self, values: '_ValueArray') -> None:
        """creates a column slice and stores a reference to the valuearray in it"""
        self.values = values
        self.property_names = []
        self.property_values = []

    def __repr__(self) -> str:
        return f"<{_utils.type_name(type(self))} object: {self.values!r}>"

    def add_property(self, name: str, values: '_ValueArray') -> None:
        """stores a named value property reference in the given column slice"""
        if name in self.property_names:
            raise SBDFError("the property with the given name already exists")
        self.property_names.append(name)
        self.property_values.append(values)

    def get_property(self, name: str) -> typing.Optional['_ValueArray']:
        """gets a value property reference with the given name"""
        if name in self.property_names:
            index = self.property_names.index(name)
            return self.property_values[index]
        return None

    def get_property_count(self) -> int:
        """gets the number of value properties in the column slice"""
        return len(self.property_names)

    def get_row_count(self) -> int:
        """gets the number of rows of the values in the column slice"""
        return self.values.row_count()

    def write(self, file: typing.BinaryIO) -> None:
        """writes a value array to file"""
        _section_write(file, _SectionTypeId.COLUMNSLICE)
        self.values.write(file)
        _write_int32(file, self.get_property_count())
        for i in range(self.get_property_count()):
            _write_string(file, self.property_names[i])
            self.property_values[i].write(file)

    @classmethod
    def read(cls, file: typing.BinaryIO) -> '_ColumnSlice':
        """reads a value array from file"""
        _section_expect(file, _SectionTypeId.COLUMNSLICE)
        values = _ValueArray.read(file)
        cslice = cls(values)
        val = _read_int32(file)
        for _ in range(val):
            name = _read_string(file)
            prop = _ValueArray.read(file)
            cslice.add_property(name, prop)
        return cslice

    @classmethod
    def skip(cls, file: typing.BinaryIO) -> None:
        """skips a value array in the file"""
        raise NotImplementedError  # sbdf_cs_skip


# Table Slices

class _TableSlice:
    def __init__(self, table_metadata: '_TableMetadata') -> None:
        """creates a table slice, storing a reference to the table metadata"""
        self.table_metadata = table_metadata
        self.columns = []

    def __repr__(self) -> str:
        return f"<{_utils.type_name(type(self))} object: {self.columns!r}>"

    def column_count(self) -> int:
        """get the number of columns in the table slice"""
        return len(self.columns)

    def add(self, column_slice: _ColumnSlice) -> None:
        """adds a column slice reference to the table slice"""
        self.columns.append(column_slice)

    def write(self, file: typing.BinaryIO) -> None:
        """writes a table slice to file"""
        _section_write(file, _SectionTypeId.TABLESLICE)
        no_columns = self.column_count()
        _write_int32(file, no_columns)
        for i in range(no_columns):
            self.columns[i].write(file)

    @staticmethod
    def write_end(file: typing.BinaryIO) -> None:
        """writes the end-of-table marker to the file"""
        _section_write(file, _SectionTypeId.TABLEEND)

    @classmethod
    def read(cls, file: typing.BinaryIO, table_metadata: '_TableMetadata',
             subset: typing.Optional[typing.List[bool]]) -> typing.Optional['_TableSlice']:
        """reads a table slice from file. returns None when the end of the table is reached"""
        val = _section_read(file)
        if val == _SectionTypeId.TABLEEND:
            return None
        if val != _SectionTypeId.TABLESLICE:
            raise SBDFError("unexpected section id")
        column_count = _read_int32(file)
        if column_count < 0:
            raise SBDFError("the number of elements is incorrect")
        if column_count != table_metadata.column_count():
            raise SBDFError("the number of the columnslice doesn't match the number of the columns of the metadata")
        tslice = cls(table_metadata)
        for i in range(column_count):
            if subset is None or subset[i]:
                tslice.add(_ColumnSlice.read(file))
            else:
                _ColumnSlice.skip(file)
        return tslice

    @classmethod
    def skip(cls, file: typing.BinaryIO, table_metadata: '_TableMetadata') -> None:
        """skips a table slice in file"""
        subset = []
        for _ in range(table_metadata.column_count()):
            subset.append(False)
        cls.read(file, table_metadata, subset)


# Column Metadata

class _ColumnMetadata:
    Property_Name = "Name"
    Property_DataType = "DataType"

    @staticmethod
    def set_values(metadata: '_Metadata', column_name: str, data_type) -> None:
        """sets the column metadata values name and data type for the previously allocated metadata head"""
        metadata.add_str(_ColumnMetadata.Property_Name, column_name)
        obj = _ValueType(data_type).as_sbdfobject()
        metadata.add(_ColumnMetadata.Property_DataType, obj)

    @staticmethod
    def get_name(metadata: '_Metadata') -> str:
        """gets the name of the column metadata"""
        name = metadata.get(_ColumnMetadata.Property_Name)
        if name.valuetype != _ValueTypeId.STRING:
            raise SBDFError("the metadata is incorrect")
        return name.data[0]

    @staticmethod
    def get_type(metadata: '_Metadata') -> '_ValueTypeId':
        """gets the value type of the column metadata"""
        obj = metadata.get(_ColumnMetadata.Property_DataType)
        if obj.valuetype != _ValueTypeId.BINARY or obj.get_count() != 1:
            raise SBDFError("the metadata is incorrect")
        return _ValueTypeId(obj.data[0][0])


# Table Metadata

class _TableMetadata:
    def __init__(self, table_metadata: '_Metadata') -> None:
        """creates the table metadata, storing a copy the given table information"""
        self.table_metadata = table_metadata
        self.table_metadata.set_immutable()
        self.column_metadata = []

    def __repr__(self) -> str:
        return f"<{_utils.type_name(type(self))} object: {self.table_metadata!r} {self.column_metadata!r}>"

    def column_count(self) -> int:
        """return the number of columns in this table"""
        return len(self.column_metadata)

    def add(self, column_metadata: '_Metadata') -> None:
        """adds column metadata to the table metadata"""
        self.column_metadata.append(column_metadata)

    def write(self, file: typing.BinaryIO) -> None:
        """writes table metadata to file"""
        _section_write(file, _SectionTypeId.TABLEMETADATA)
        _write_int32(file, self.table_metadata.count())
        for i in range(self.table_metadata.count()):
            _write_string(file, self.table_metadata.names[i])
            _write_int8(file, self.table_metadata.values[i].valuetype)
            _write_optional_value(file, self.table_metadata.values[i])
            _write_optional_value(file, self.table_metadata.default_values[i])
        _write_int32(file, self.column_count())
        # fold duplicate values
        cm_default = {}
        cm_types = {}
        for cmeta in self.column_metadata:
            for i in range(cmeta.count()):
                cm_name = cmeta.names[i]
                if cm_name in cm_default and cm_default[cm_name] != cmeta.default_values[i]:
                    raise SBDFError("the metadata is incorrect")
                cm_default[cm_name] = cmeta.default_values[i]
                if cmeta.values[i] is not None:
                    cm_type = cmeta.values[i].valuetype
                else:
                    cm_type = cmeta.default_values[i].valuetype
                if cm_name in cm_types and cm_types[cm_name] != cm_type:
                    raise SBDFError("the metadata is incorrect")
                cm_types[cm_name] = cm_type
        _write_int32(file, len(cm_default))
        # write names, data types, and default values
        for i in sorted(cm_default):
            _write_string(file, i)
            _ValueType(cm_types[i]).write(file)
            _write_optional_value(file, cm_default[i])
        # write column values
        for cmeta in self.column_metadata:
            for i in sorted(cm_default):
                val = cmeta.get(i)
                _write_optional_value(file, val)

    @classmethod
    def read(cls, file: typing.BinaryIO) -> '_TableMetadata':
        """reads table metadata from file"""
        _section_expect(file, _SectionTypeId.TABLEMETADATA)
        metadata_count = _read_int32(file)
        if metadata_count < 0:
            raise SBDFError("the number of elements is incorrect")
        metadata = _Metadata()
        for _ in range(metadata_count):
            name = _read_string(file)
            valtype = _ValueType.read(file)
            value_present = _read_int8(file)
            value = _SbdfObject.read(file, valtype) if value_present else None
            value_present = _read_int8(file)
            default_value = _SbdfObject.read(file, valtype) if value_present else None
            metadata.add(name, value, default_value)
        tmeta = _TableMetadata(metadata)

        column_count = _read_int32(file)
        metadata_count = _read_int32(file)
        md_name = []
        md_type = []
        md_default = []
        for i in range(metadata_count):
            md_name.append(_read_string(file))
            md_type.append(_ValueType.read(file))
            value_present = _read_int8(file)
            md_default.append(_SbdfObject.read(file, md_type[i]) if value_present else None)
        for i in range(column_count):
            metadata = _Metadata()
            for j in range(metadata_count):
                value_present = _read_int8(file)
                if value_present:
                    value = _SbdfObject.read(file, md_type[j])
                    metadata.add(md_name[j], value, md_default[j])
            tmeta.add(metadata)
        return tmeta


# Metadata

class _Metadata:
    def __init__(self) -> None:
        """creates an empty metadata structure"""
        self.modifiable = True
        self.names = []
        self.values = []
        self.default_values = []

    def __repr__(self) -> str:
        return f"<{_utils.type_name(type(self))} object: {self.names!r} -> {self.values!r}>"

    def add_str(self, name: str, value: str, default_value: str = None) -> None:
        """adds a named string metadata value and default value to out"""
        val = _SbdfObject(_ValueTypeId.STRING, [str(value)])
        if default_value is not None:
            default = _SbdfObject(_ValueTypeId.STRING, [str(default_value)])
        else:
            default = None
        self.add(name, val, default)

    def add_int(self, name: str, value: int, default_value: int = None) -> None:
        """adds a named integer metadata value and default value to out"""
        val = _SbdfObject(_ValueTypeId.INT, [value])
        if default_value is not None:
            default = _SbdfObject(_ValueTypeId.INT, [default_value])
        else:
            default = None
        self.add(name, val, default)

    def add(self, name: str, value: '_SbdfObject', default_value: '_SbdfObject' = None) -> None:
        """adds a named metadata value and default value to out"""
        if not self.modifiable:
            raise SBDFError("the metadata is readonly and may not be modified")
        if default_value is not None and value.valuetype != default_value.valuetype:
            raise SBDFError("the valuetypes of the arguments must be equal")
        if name in self.names:
            raise SBDFError("the metadata with the given name already exists")
        self.names.append(name)
        self.values.append(value)
        self.default_values.append(default_value)

    def remove(self, name: str) -> None:
        """removes the named metadata value from out"""
        if not self.modifiable:
            raise SBDFError("the metadata is readonly and may not be modified")
        if name in self.names:
            index = self.names.index(name)
            del self.names[index]
            del self.values[index]
            del self.default_values[index]

    def get(self, name: str) -> typing.Optional['_SbdfObject']:
        """gets a copy of the named metadata value"""
        if name in self.names:
            index = self.names.index(name)
            return self.values[index]
        return None

    def get_default(self, name: str) -> typing.Optional['_SbdfObject']:
        """gets a copy of the named default metadata value"""
        if name in self.names:
            index = self.names.index(name)
            return self.default_values[index]
        return None

    def count(self) -> int:
        """returns the number of metadata entries pointed to by head"""
        return len(self.names)

    def exists(self, name: str) -> bool:
        """returns True if the named metadata exists. False is returned if the metadata doesn't exist"""
        return name in self.names

    def set_immutable(self) -> None:
        """sets the metadata immutable so that it may not be modified by subsequent operations"""
        self.modifiable = False


# Objects

class _SbdfObject:
    def __init__(self, valuetype: '_ValueTypeId', data: typing.List) -> None:
        self.valuetype = valuetype
        self.data = data

    def __repr__(self) -> str:
        return f"<{_utils.type_name(type(self))} object ({self.valuetype!r}): {self.data!r}>"

    def get_count(self) -> int:
        """get the number of items in the object"""
        return len(self.data)

    def write_array(self, file: typing.BinaryIO) -> None:
        """writes the array information to the specified file. valuetype information is not written"""
        count = len(self.data)
        _write_int32(file, count)
        self._write_n(file, count, True)

    def write(self, file: typing.BinaryIO) -> None:
        """writes the object to the specified file. valuetype information is not written"""
        self._write_n(file, 1, False)

    def _write_n(self, file: typing.BinaryIO, n: int, packed: bool) -> None:
        valtype = _ValueType(self.valuetype)
        if valtype.is_array():
            byte_size = 0
            if packed:
                saved_bytes = []
                for i in range(n):
                    saved_bytes.append(valtype.to_bytes(self.data[i]))
                    length = len(saved_bytes[i])
                    byte_size += _get_7bit_packed_length(length) + length
                _write_int32(file, byte_size)
                for i in range(n):
                    length = len(saved_bytes[i])
                    _write_7bit_packed_int32(file, length)
                    if length:
                        _write_bytes(file, saved_bytes[i])
            else:
                for i in range(n):
                    valtype_bytes = valtype.to_bytes(self.data[i])
                    length = len(valtype_bytes)
                    _write_int32(file, length)
                    if length:
                        _write_bytes(file, valtype_bytes)
        else:
            size = valtype.get_packed_size()
            if size is None:
                raise SBDFError("unknown typeid")
            for i in range(n):
                valtype_bytes = valtype.to_bytes(self.data[i])
                _write_bytes(file, valtype_bytes)

    @classmethod
    def read_array(cls, file: typing.BinaryIO, valuetype: '_ValueType') -> '_SbdfObject':
        """reads an array object with the given value type from the specified file"""
        count = _read_int32(file)
        return cls._read_n(file, count, valuetype, True)

    @classmethod
    def read(cls, file: typing.BinaryIO, valuetype: '_ValueType') -> '_SbdfObject':
        """reads an object with the given valuetype from the file."""
        return cls._read_n(file, 1, valuetype, False)

    @classmethod
    def _read_n(cls, file: typing.BinaryIO, n: int, valuetype: '_ValueType', packed: bool) -> '_SbdfObject':
        data = []
        if valuetype.is_array():
            # read byte size and ignore it
            if packed:
                _read_int32(file)
            for _ in range(n):
                if packed:
                    length = _read_7bit_packed_int32(file)
                else:
                    length = _read_int32(file)
                if length < 0:
                    raise SBDFError("the number of elements is incorrect")
                dest = _read_bytes(file, length)
                data.append(valuetype.to_python(dest))
        else:
            size = valuetype.get_packed_size()
            if size is None:
                raise SBDFError("unknown typeid")
            for _ in range(n):
                dest = _read_bytes(file, size)
                data.append(valuetype.to_python(dest))
        return cls(valuetype.type_id, data)

    @classmethod
    def skip_array(cls, file: typing.BinaryIO, valuetype: '_ValueType') -> None:
        """skips an array with the given valuetype"""
        raise NotImplementedError  # sbdf_obj_skip_arr

    @classmethod
    def skip(cls, file: typing.BinaryIO, valuetype: '_ValueType') -> None:
        """skips an object with the given valuetype"""
        raise NotImplementedError  # sbdf_obj_skip


# Value arrays

class _ValueArrayEncoding(enum.IntEnum):
    PLAIN_ARRAY = 0x1
    RUN_LENGTH = 0x2
    BIT_ARRAY = 0x3


class _ValueArray:
    def __init__(self, array_encoding: _ValueArrayEncoding, array: typing.Optional[_SbdfObject]) -> None:
        """creates a value array from the specified values"""
        self.encoding = _ValueArrayEncoding(array_encoding)
        self.val1 = None
        self.obj1 = None
        self.obj2 = None
        self.valuetype = None
        if array is None:
            pass
        elif array_encoding == _ValueArrayEncoding.PLAIN_ARRAY:
            self._create_plain(array)
        elif array_encoding == _ValueArrayEncoding.RUN_LENGTH:
            self._create_rle(array)
        elif array_encoding == _ValueArrayEncoding.BIT_ARRAY:
            self._create_bit(array)
        else:
            raise SBDFError("unknown valuearray encoding")

    def _create_plain(self, array: _SbdfObject) -> None:
        self.valuetype = array.valuetype
        self.obj1 = array

    def _create_rle(self, array: _SbdfObject) -> None:
        raise NotImplementedError  # sbdf_va_create_rle

    def _create_bit(self, array: _SbdfObject) -> None:
        self.valuetype = _ValueTypeId.BOOL
        self.val1 = array.get_count()
        bits = bitstring.BitArray(array.data)
        while bits.len % 8 != 0:
            bits.append('0b0')
        self.obj1 = _SbdfObject(_ValueTypeId.BINARY, [bits.bytes])

    def __repr__(self) -> str:
        if self.encoding == _ValueArrayEncoding.PLAIN_ARRAY:
            arr = self.obj1
        elif self.encoding == _ValueArrayEncoding.RUN_LENGTH:
            arr = "..."
        elif self.encoding == _ValueArrayEncoding.BIT_ARRAY:
            arr = self.obj1.data[0]
        else:
            arr = "unknown encoding"
        return f"<{_utils.type_name(type(self))} object ({self.encoding!r}): {arr!r}>"

    def get_values(self) -> _SbdfObject:
        """extracts the values from the array"""
        if self.encoding == _ValueArrayEncoding.PLAIN_ARRAY:
            return self.obj1
        if self.encoding == _ValueArrayEncoding.RUN_LENGTH:
            return self._get_rle()
        if self.encoding == _ValueArrayEncoding.BIT_ARRAY:
            return self._get_bit()
        raise SBDFError("unknown valuearray encoding")

    def _get_rle(self) -> _SbdfObject:
        raise NotImplementedError  # sbdf_get_rle_values

    def _get_bit(self) -> _SbdfObject:
        obj = _SbdfObject(_ValueTypeId.BOOL, [])
        bits = bitstring.BitArray(bytes=self.obj1.data[0], length=self.val1)
        for i in bits:
            obj.data.append(i)
        return obj

    def row_count(self) -> int:
        """returns the number of rows stored in the value array"""
        if self.encoding == _ValueArrayEncoding.PLAIN_ARRAY:
            return self.obj1.get_count()
        if self.encoding == _ValueArrayEncoding.RUN_LENGTH:
            return self.val1
        if self.encoding == _ValueArrayEncoding.BIT_ARRAY:
            return self.val1
        raise SBDFError("unknown valuearray encoding")

    def write(self, file: typing.BinaryIO) -> None:
        """writes the value array to the current file position"""
        _write_int8(file, self.encoding)
        _ValueType(self.valuetype).write(file)
        if self.encoding == _ValueArrayEncoding.PLAIN_ARRAY:
            self.obj1.write_array(file)
        elif self.encoding == _ValueArrayEncoding.RUN_LENGTH:
            _write_int32(file, self.val1)
            self.obj1.write_array(file)
            self.obj2.write_array(file)
        elif self.encoding == _ValueArrayEncoding.BIT_ARRAY:
            _write_int32(file, self.val1)
            _write_bytes(file, self.obj1.data[0])
        else:
            raise SBDFError("unknown valuearray encoding")

    @classmethod
    def read(cls, file: typing.BinaryIO) -> '_ValueArray':
        """reads the value array from the current file position"""
        encoding = _read_int8(file)
        valtype = _ValueType.read(file)
        handle = cls(_ValueArrayEncoding(encoding), None)
        handle.valuetype = valtype
        if encoding == _ValueArrayEncoding.PLAIN_ARRAY:
            handle.obj1 = _SbdfObject.read_array(file, valtype)
        elif encoding == _ValueArrayEncoding.RUN_LENGTH:
            handle.val1 = _read_int32(file)
            handle.obj1 = _SbdfObject.read_array(file, _ValueType(_ValueTypeId.INTERNAL_BYTE))
            handle.obj2 = _SbdfObject.read_array(file, valtype)
        elif encoding == _ValueArrayEncoding.BIT_ARRAY:
            val = _read_int32(file)
            handle.val1 = val
            packed_size = val // 8 + (1 if val % 8 else 0)
            bits = _read_bytes(file, packed_size)
            handle.obj1 = _SbdfObject(_ValueTypeId.BINARY, [bits])
        else:
            raise SBDFError("unknown valuearray encoding")
        return handle

    @classmethod
    def skip(cls, file: typing.BinaryIO) -> None:
        """skips the value array at the current file position"""
        encoding = _read_int8(file)
        valtype = _ValueType.read(file)
        if encoding == _ValueArrayEncoding.PLAIN_ARRAY:
            _SbdfObject.skip_array(file, valtype)
        elif encoding == _ValueArrayEncoding.RUN_LENGTH:
            _SbdfObject.skip_array(file, _ValueType(_ValueTypeId.INTERNAL_BYTE))
            _SbdfObject.skip_array(file, valtype)
        elif encoding == _ValueArrayEncoding.BIT_ARRAY:
            val = _read_int32(file)
            packed_size = val // 8 + (1 if val % 8 else 0)
            file.seek(packed_size, 1)
        else:
            raise SBDFError("unknown valuearray encoding")


# Value types

class _ValueTypeId(enum.IntEnum):
    UNKNOWN = 0x00
    BOOL = 0x01      # C type is char
    INT = 0x02       # C type is 32-bit int
    LONG = 0x03      # C type is 64-bit int
    FLOAT = 0x04     # C type is float
    DOUBLE = 0x05    # C type is double
    DATETIME = 0x06  # C representation is milliseconds since 01/01/01, 00:00:00, stored in a 64-bit int
    DATE = 0x07      # C representation is milliseconds since 01/01/01, 00:00:00, stored in a 64-bit int
    TIME = 0x08      # C representation is milliseconds since 01/01/01, 00:00:00, stored in a 64-bit int
    TIMESPAN = 0x09  # C representation is milliseconds, stored in a 64-bit int
    STRING = 0x0a    # C representation is char-ptr
    BINARY = 0x0c    # C representation is void-ptr
    DECIMAL = 0x0d   # C representation is IEEE754 decimal128 Binary Integer Decimals
    INTERNAL_BYTE = 0xfe

    def to_typename_string(self) -> str:
        """convert this valuetype id to the type name used by Spotfire"""
        return {
            _ValueTypeId.BOOL: "Boolean",
            _ValueTypeId.INT: "Integer",
            _ValueTypeId.LONG: "LongInteger",
            _ValueTypeId.FLOAT: "SingleReal",
            _ValueTypeId.DOUBLE: "Real",
            _ValueTypeId.DATETIME: "DateTime",
            _ValueTypeId.DATE: "Date",
            _ValueTypeId.TIME: "Time",
            _ValueTypeId.TIMESPAN: "TimeSpan",
            _ValueTypeId.STRING: "String",
            _ValueTypeId.BINARY: "Binary",
            _ValueTypeId.DECIMAL: "Currency"
        }.get(self, "unknown")

    def to_dtype_name(self) -> str:
        """convert this valuetype id to the dtype name used by Pandas"""
        return {
            _ValueTypeId.INT: "Int32",
            _ValueTypeId.LONG: "Int64",
            _ValueTypeId.FLOAT: "float32",
            _ValueTypeId.DOUBLE: "float64"
        }.get(self, "object")

    @staticmethod
    def infer_from_type(values, value_description: str) -> '_ValueTypeId':
        """determine the proper valuetype id from the Python types in a column"""
        # Remove any None (or other none-ish things) from values
        vals = [x for x in values if not pd.isnull(x)]
        # Check if any values remain
        if not vals:
            raise SBDFError(f"cannot determine type for {value_description}; all values are missing")
        # Check to make sure only one type remains
        vals_type = type(vals[0])
        if not all(isinstance(i, vals_type) for i in vals):
            raise SBDFError(f"types in {value_description} do not match")
        # Determine the right type id
        typeid = {
            bool: _ValueTypeId.BOOL,
            np.int32: _ValueTypeId.INT,
            int: _ValueTypeId.LONG,
            np.int64: _ValueTypeId.LONG,
            np.float32: _ValueTypeId.FLOAT,
            float: _ValueTypeId.DOUBLE,
            np.float64: _ValueTypeId.DOUBLE,
            datetime.datetime: _ValueTypeId.DATETIME,
            pd.Timestamp: _ValueTypeId.DATETIME,
            datetime.date: _ValueTypeId.DATE,
            datetime.time: _ValueTypeId.TIME,
            datetime.timedelta: _ValueTypeId.TIMESPAN,
            pd.Timedelta: _ValueTypeId.TIMESPAN,
            str: _ValueTypeId.STRING,
            bytes: _ValueTypeId.BINARY,
            decimal.Decimal: _ValueTypeId.DECIMAL,
        }.get(vals_type, None)
        if typeid is None:
            raise SBDFError(f"unknown type '{_utils.type_name(vals_type)}' in {value_description}")
        return typeid

    @staticmethod
    def infer_from_dtype(series: pd.Series, series_description: str) -> '_ValueTypeId':
        """determine the proper valuetype id from the Pandas dtype of a series"""
        dtype = series.dtype.name
        if dtype == "object":
            return _ValueTypeId.infer_from_type(series.tolist(), series_description)
        if dtype == "category":
            return _ValueTypeId.infer_from_dtype(series.astype(series.cat.categories.dtype), series_description)
        typeid = {
            "bool": _ValueTypeId.BOOL,
            "int32": _ValueTypeId.INT,
            "Int32": _ValueTypeId.INT,
            "int64": _ValueTypeId.LONG,
            "Int64": _ValueTypeId.LONG,
            "float32": _ValueTypeId.FLOAT,
            "float64": _ValueTypeId.DOUBLE,
            "datetime64[ns]": _ValueTypeId.DATETIME,
            "timedelta64[ns]": _ValueTypeId.TIMESPAN,
            "string": _ValueTypeId.STRING,
        }.get(dtype, None)
        if typeid is None:
            raise SBDFError(f"unknown dtype '{dtype}' in {series_description}")
        return typeid


class _ValueType:
    def __init__(self, type_id: int) -> None:
        self.type_id = _ValueTypeId(type_id)

    def __repr__(self) -> str:
        return str(self.type_id)

    def __cmp__(self, other: '_ValueType') -> int:
        return self.type_id - other.type_id

    def write(self, file: typing.BinaryIO) -> None:
        """writes a valuetype to the current file position"""
        _write_int8(file, self.type_id)

    @classmethod
    def read(cls, file: typing.BinaryIO) -> '_ValueType':
        """reads a valuetype from the current file position"""
        return cls(_read_int8(file))

    def is_array(self) -> bool:
        """determines if this valuetype is an array type index"""
        return self.type_id in (_ValueTypeId.STRING, _ValueTypeId.BINARY)

    def get_packed_size(self) -> int:
        """returns the packed byte size (on disk) of a valuetype"""
        return {
            _ValueTypeId.BOOL: 1,
            _ValueTypeId.INT: 4,
            _ValueTypeId.LONG: 8,
            _ValueTypeId.FLOAT: 4,
            _ValueTypeId.DOUBLE: 8,
            _ValueTypeId.DATETIME: 8,
            _ValueTypeId.DATE: 8,
            _ValueTypeId.TIME: 8,
            _ValueTypeId.TIMESPAN: 8,
            _ValueTypeId.STRING: 0,  # size is dynamic
            _ValueTypeId.BINARY: 0,  # size is dynamic
            _ValueTypeId.DECIMAL: 16
        }.get(self.type_id, None)

    _DATETIME_EPOCH = datetime.datetime(1, 1, 1)
    _DECIMAL_EXPONENT_BIAS = 12320

    @staticmethod
    def _to_python_bool(data: bytes) -> bool:
        return struct.unpack("?", data)[0]

    @staticmethod
    def _to_python_int(data: bytes) -> int:
        return struct.unpack("<i", data)[0]

    @staticmethod
    def _to_python_long(data: bytes) -> int:
        return struct.unpack("<q", data)[0]

    @staticmethod
    def _to_python_float(data: bytes) -> float:
        return struct.unpack("<f", data)[0]

    @staticmethod
    def _to_python_double(data: bytes) -> float:
        return struct.unpack("<d", data)[0]

    @staticmethod
    def _to_python_datetime(data: bytes) -> datetime.datetime:
        timestamp = struct.unpack("<q", data)[0]
        return _ValueType._DATETIME_EPOCH + datetime.timedelta(milliseconds=timestamp)

    @staticmethod
    def _to_python_date(data: bytes) -> datetime.date:
        timestamp = struct.unpack("<q", data)[0]
        date = _ValueType._DATETIME_EPOCH + datetime.timedelta(milliseconds=timestamp)
        return date.date()

    @staticmethod
    def _to_python_time(data: bytes) -> datetime.time:
        timestamp = struct.unpack("<q", data)[0]
        date = _ValueType._DATETIME_EPOCH + datetime.timedelta(milliseconds=timestamp)
        return date.timetz()

    @staticmethod
    def _to_python_timespan(data: bytes) -> datetime.timedelta:
        timespan = struct.unpack("<q", data)[0]
        return datetime.timedelta(milliseconds=timespan)

    @staticmethod
    def _to_python_string(data: bytes) -> str:
        return data.decode("utf-8")

    @staticmethod
    def _to_python_binary(data: bytes) -> bytes:
        return data

    @staticmethod
    def _to_python_decimal(data: bytes) -> decimal.Decimal:
        bits = bitstring.BitArray(bytes=data)
        # pylint: disable=unbalanced-tuple-unpacking
        coefficient, biased_exponent_bits_high, sign_bit, biased_exponent_bits_low = \
            bits.unpack('uintle:96,pad:17,bits:7,bool,bits:7')
        # un-bias the exponent
        biased_exponent_bits = bitstring.BitArray('0b00')
        biased_exponent_bits.append(biased_exponent_bits_high)
        biased_exponent_bits.append(biased_exponent_bits_low)
        exponent = biased_exponent_bits.uintle - _ValueType._DECIMAL_EXPONENT_BIAS
        # break up the coefficient into its digits
        digits = []
        while coefficient != 0:
            digits.insert(0, coefficient % 10)
            coefficient //= 10
        # construct the decimal value
        return decimal.Decimal((1 if sign_bit else 0, tuple(digits), exponent))

    def to_python(self, data: bytes) -> typing.Any:
        """return a Python representation of the raw data"""
        return getattr(self, "_to_python_" + self.type_id.name.lower(), lambda x: None)(data)

    @staticmethod
    def _to_bytes_bool(obj: bool) -> bytes:
        return struct.pack("?", obj)

    @staticmethod
    def _to_bytes_int(obj: int) -> bytes:
        return struct.pack("<i", obj)

    @staticmethod
    def _to_bytes_long(obj: int) -> bytes:
        return struct.pack("<q", obj)

    @staticmethod
    def _to_bytes_float(obj: float) -> bytes:
        return struct.pack("<f", obj)

    @staticmethod
    def _to_bytes_double(obj: float) -> bytes:
        return struct.pack("<d", obj)

    @staticmethod
    def _to_bytes_datetime(obj: datetime.datetime) -> bytes:
        if isinstance(obj, pd.Timestamp):
            obj_dt = obj.to_pydatetime()
        else:
            obj_dt = obj
        td_after_epoch = obj_dt - _ValueType._DATETIME_EPOCH
        timespan = int(td_after_epoch / datetime.timedelta(milliseconds=1))
        return struct.pack("<q", timespan)

    @staticmethod
    def _to_bytes_date(obj: datetime.date) -> bytes:
        td_after_epoch = obj - _ValueType._DATETIME_EPOCH.date()
        timespan = int(td_after_epoch / datetime.timedelta(milliseconds=1))
        return struct.pack("<q", timespan)

    @staticmethod
    def _to_bytes_time(obj: datetime.time) -> bytes:
        obj_td = datetime.datetime.combine(datetime.datetime.min, obj) - datetime.datetime.min
        timestamp = obj_td // datetime.timedelta(milliseconds=1)
        return struct.pack("<q", timestamp)

    @staticmethod
    def _to_bytes_timespan(obj: datetime.timedelta) -> bytes:
        if isinstance(obj, pd.Timedelta):
            obj_td = obj.to_pytimedelta()
        elif pd.isnull(obj):
            obj_td = datetime.timedelta(0)
        else:
            obj_td = obj
        timespan = int(obj_td / datetime.timedelta(milliseconds=1))
        return struct.pack("<q", timespan)

    @staticmethod
    def _to_bytes_string(obj: str) -> bytes:
        return obj.encode("utf-8")

    @staticmethod
    def _to_bytes_binary(obj: bytes) -> bytes:
        return obj

    @staticmethod
    def _to_bytes_decimal(obj: decimal.Decimal) -> bytes:
        dec = obj.as_tuple()
        # bias the exponent and convert it to a 14-bit value
        biased_exponent = dec.exponent + _ValueType._DECIMAL_EXPONENT_BIAS
        biased_exponent_bits = bitstring.pack('uintle:16', biased_exponent)
        biased_exponent_bits_high = biased_exponent_bits[2:9]
        biased_exponent_bits_low = biased_exponent_bits[9:16]
        # combine the digits of the coefficient
        coefficient = 0
        for i in dec.digits:
            coefficient *= 10
            coefficient += i
        # construct the binary128 value
        return bitstring.pack('uintle:96,pad:17,bits:7,bool,bits:7', coefficient, biased_exponent_bits_high,
                              dec.sign, biased_exponent_bits_low).bytes

    def to_bytes(self, obj: typing.Any) -> bytes:
        """return a SBDF representation of the Python objects"""
        try:
            return getattr(self, "_to_bytes_" + self.type_id.name.lower(), lambda x: None)(obj)
        except (struct.error, bitstring.CreationError, UnicodeError) as exc:
            raise SBDFError(f"cannot convert '{obj}' to Spotfire {self.type_id.to_typename_string()} \
            type; value is outside representable range") from exc

    def missing_value(self) -> typing.Any:
        """return a missing value appropriate for the value type"""
        val = None
        if self.type_id == _ValueTypeId.BOOL:
            val = False
        elif self.type_id == _ValueTypeId.INT:
            val = 0
        elif self.type_id == _ValueTypeId.LONG:
            val = 0
        elif self.type_id == _ValueTypeId.FLOAT:
            val = 0.0
        elif self.type_id == _ValueTypeId.DOUBLE:
            val = 0.0
        elif self.type_id == _ValueTypeId.DATETIME:
            val = self._DATETIME_EPOCH
        elif self.type_id == _ValueTypeId.DATE:
            val = self._DATETIME_EPOCH.date()
        elif self.type_id == _ValueTypeId.TIME:
            val = self._DATETIME_EPOCH.time()
        elif self.type_id == _ValueTypeId.TIMESPAN:
            val = datetime.timedelta(0)
        elif self.type_id == _ValueTypeId.STRING:
            val = ""
        elif self.type_id == _ValueTypeId.BINARY:
            val = b""
        elif self.type_id == _ValueTypeId.DECIMAL:
            val = decimal.Decimal()
        return val

    def as_sbdfobject(self) -> _SbdfObject:
        """return an SbdfObject that describes this ValueType"""
        return _SbdfObject(_ValueTypeId.BINARY, [struct.pack(">B", self.type_id)])


# Sections

class _SectionTypeId(enum.IntEnum):
    # An unknown section type.
    UNKNOWN = 0x0

    # A file header section.
    FILEHEADER = 0x1

    # A table metadata section, marking the beginning of a complete table.
    TABLEMETADATA = 0x2

    # A table slice section.
    TABLESLICE = 0x3

    # A column slice section.
    COLUMNSLICE = 0x4

    # Marks the end of a complete data table.
    TABLEEND = 0x5


def _section_write(file: typing.BinaryIO, section_id: _SectionTypeId) -> None:
    """writes the given section type id to file"""
    _write_int8(file, 0xdf)
    _write_int8(file, 0x5b)
    _write_int8(file, section_id)


def _section_read(file: typing.BinaryIO) -> int:
    """reads the given section type id from file"""
    val = _read_int8(file)
    if val != 0xdf:
        raise SBDFError("the SBDF magic number wasn't found")
    val = _read_int8(file)
    if val != 0x5b:
        raise SBDFError("the SBDF magic number wasn't found")
    return _read_int8(file)


def _section_expect(file: typing.BinaryIO, section_id: _SectionTypeId) -> None:
    """reads the section type id from file. raises error if different from passed id"""
    val = _section_read(file)
    if val != section_id:
        raise SBDFError("unexpected section id")


# Internals

def _write_bytes(file: typing.BinaryIO, value: bytes) -> int:
    if "b" not in file.mode:
        raise SBDFError("file not opened as binary")
    written = file.write(value)
    if written != len(value):
        raise SBDFError("i/o error")
    return written


def _read_bytes(file: typing.BinaryIO, n: int) -> bytes:
    if "b" not in file.mode:
        raise SBDFError("file not opened as binary")
    bytes_read = file.read(n)
    if len(bytes_read) != n:
        raise SBDFError("i/o error")
    return bytes_read


def _write_int8(file: typing.BinaryIO, value: int) -> None:
    to_write = struct.pack("B", value)
    _write_bytes(file, to_write)


def _read_int8(file: typing.BinaryIO) -> int:
    bytes_read = _read_bytes(file, 1)
    return struct.unpack("B", bytes_read)[0]


def _write_int32(file: typing.BinaryIO, value: int) -> None:
    to_write = struct.pack("<i", value)
    _write_bytes(file, to_write)


def _read_int32(file: typing.BinaryIO) -> int:
    bytes_read = _read_bytes(file, 4)
    return struct.unpack("<i", bytes_read)[0]


def _write_string(file: typing.BinaryIO, string: str) -> None:
    to_write = string.encode("utf-8")
    _write_int32(file, len(to_write))
    _write_bytes(file, to_write)


def _read_string(file: typing.BinaryIO) -> str:
    size = _read_int32(file)
    if size < 0:
        raise SBDFError("the number of elements is incorrect")
    bytes_read = _read_bytes(file, size)
    return bytes_read.decode("utf-8")


def _skip_string(file: typing.BinaryIO) -> None:
    size = _read_int32(file)
    if size < 0:
        raise SBDFError("the number of elements is incorrect")
    file.seek(size, 1)


def _write_7bit_packed_int32(file: typing.BinaryIO, value: int) -> None:
    while True:
        uch = bytearray(1)
        uch[0] = value & 0x7f
        if value > 0x7f:
            uch[0] |= 0x80
        if _write_bytes(file, uch) != 1:
            raise SBDFError("i/o error")
        if value > 0x7f:
            value >>= 7
        else:
            break


def _read_7bit_packed_int32(file: typing.BinaryIO) -> int:
    value = 0
    shift = 0
    while True:
        byte_read = _read_bytes(file, 1)
        byte_as_int = byte_read[0]
        value |= (byte_as_int & 0x7f) << shift
        if (byte_as_int & 0x80) == 0x80:
            shift += 7
        else:
            break
    return value


def _get_7bit_packed_length(value: int) -> int:
    length = 5
    if value < (1 << 7):
        length = 1
    elif value < (1 << 14):
        length = 2
    elif value < (1 << 21):
        length = 3
    elif value < (1 << 28):
        length = 4
    return length


def _write_optional_value(file: typing.BinaryIO, value) -> None:
    if value is None:
        _write_int8(file, 0)
    else:
        _write_int8(file, 1)
        value.write(file)


if gpd is not None:
    def _geo_data_frame_to_data_frame(gdf: gpd.GeoDataFrame,
                                      table_metadata: typing.Dict[str, typing.Any],
                                      column_metadata: typing.Dict[str, typing.Dict[str, typing.Any]]) -> \
            pd.DataFrame:
        if not isinstance(gdf.get("geometry")[0], shp_geom.BaseGeometry):
            return gdf

        # Write geocoding column metadata
        cols = ["XMin", "XMax", "YMin", "YMax", "XCenter", "YCenter", "Geometry"]
        metacols = list(gdf.columns.drop('geometry'))
        metacols.extend(cols)
        for col in metacols:
            if col not in column_metadata.keys():
                column_metadata[col] = {}
        for col in cols:
            column_metadata[col]["MapChart.ColumnTypeId"] = col
        column_metadata["Geometry"]["ContentType"] = "application/x-wkb"

        # Convert to dataframe
        geom = [shapely.wkb.dumps(x) for x in gdf["geometry"]]
        geom_series = gdf["geometry"]
        gdf = gdf.drop(columns="geometry")
        names = gdf.keys()
        dframe = pd.DataFrame(gdf.to_numpy(), columns=names)

        # Add spotfire geocoding columns
        dframe["Geometry"] = pd.Series(geom)
        xmin, ymin, xmax, ymax = zip(*[x.bounds for x in geom_series])
        xcen, ycen = zip(*[x.centroid.coords[:][0] for x in geom_series])
        dframe["XMin"] = pd.Series(xmin)
        dframe["XMax"] = pd.Series(xmax)
        dframe["YMin"] = pd.Series(ymin)
        dframe["YMax"] = pd.Series(ymax)
        dframe["XCenter"] = pd.Series(xcen)
        dframe["YCenter"] = pd.Series(ycen)

        # Write geocoding table metadata
        table_metadata["MapChart.IsGeocodingTable"] = True
        if not table_metadata.get("MapChart.IsGeocodingEnabled"):
            table_metadata["MapChart.IsGeocodingEnabled"] = True
        if all(isinstance(x, shapely.geometry.Point) for x in geom_series):
            table_metadata["MapChart.GeometryType"] = "Point"
        elif all(isinstance(x, shapely.geometry.LineString) for x in geom_series) or \
                all(isinstance(x, shapely.geometry.LinearRing) for x in geom_series):
            table_metadata["MapChart.GeometryType"] = "Line"
        elif all(isinstance(x, shapely.geometry.Polygon) for x in geom_series):
            table_metadata["MapChart.GeometryType"] = "Polygon"
        else:
            raise SBDFError("cannot convert collections of Shapely objects")
        if gdf.crs is not None:
            try:
                table_metadata["MapChart.GeographicCrs"] = gdf.crs.to_string()
            except AttributeError:
                # GeoPandas <= 0.6.3 compatibility
                if gdf.crs.startswith("+init="):
                    gdf.crs = gdf.crs[6:]
                table_metadata["MapChart.GeographicCrs"] = gdf.crs
        return dframe


    def _data_frame_to_geo_data_frame(dframe: pd.DataFrame,
                                      table_metadata: typing.Dict[str, typing.Any]) -> gpd.GeoDataFrame:
        if not isinstance(dframe.get('Geometry')[0], bytes):
            raise SBDFError("cannot convert to geoDataFrame")

        # Convert to geodataframe
        geom = [shapely.wkb.loads(x) for x in dframe["Geometry"]]
        dframe = dframe.drop(columns="Geometry")
        gdf = gpd.GeoDataFrame(dframe, geometry=geom)

        # Decide what CRS to use
        if "MapChart.GeographicCrs" in table_metadata.keys() and table_metadata["MapChart.GeographicCrs"] != "":
            proj = table_metadata["MapChart.GeographicCrs"][0]
            gdf.crs = proj
            try:
                # GeoPandas <= 0.6.3 compatibility
                if not gdf.crs.startswith("+init="):
                    gdf.crs = "+init=" + proj
            except AttributeError:
                pass

        return gdf

if matplotlib is not None:
    def _pyplot_to_binary(fig: matplotlib.figure.Figure) -> bytes:
        fig.set_canvas(matplotlib.pyplot.gcf().canvas)
        with tempfile.NamedTemporaryFile(suffix=".png") as file:
            fig.savefig(file, format="png")
            return _image_file_to_binary(file)

if seaborn is not None:
    def _seaborn_plot_to_binary(plot: seaborn.axisgrid.Grid) -> bytes:
        with tempfile.NamedTemporaryFile(suffix=".png") as file:
            plot.savefig(file)
            return _image_file_to_binary(file)

if PIL is not None:
    def _pil_image_to_binary(img: PIL.Image.Image) -> bytes:
        with tempfile.NamedTemporaryFile(suffix=".png") as file:
            img.save(file, "png")
            return _image_file_to_binary(file)


def _image_file_to_binary(file: tempfile.NamedTemporaryFile) -> bytes:
    file.seek(0)
    return file.read()



# Copyright © 2021. TIBCO Software Inc.
# This file is subject to the license terms contained
# in the license file that is distributed with this file.

"""Automatically manage the protocol between Spotfire and a data function written in Python."""

import io
import os.path
import pprint
import sys
import traceback
import typing
import re

from spotfire import sbdf, _utils


def _bad_string(str_: typing.Any) -> bool:
    return not isinstance(str_, str)


class _OutputCapture:
    def __init__(self) -> None:
        self._entered = False
        self._stdout = io.StringIO()
        self._stderr = io.StringIO()
        self._old_stdout = None
        self._old_stderr = None

    def __enter__(self):
        if self._entered:
            raise ValueError("cannot reenter output capture")
        self._entered = True
        self._old_stdout = sys.stdout
        self._old_stderr = sys.stderr
        sys.stdout = self._stdout
        sys.stderr = self._stderr
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self._entered:
            raise ValueError("cannot exit unentered output capture")
        self._entered = False
        sys.stdout = self._old_stdout
        sys.stderr = self._old_stderr
        self._old_stdout = None
        self._old_stderr = None

    def get_stdout(self) -> typing.Optional[str]:
        """Return the captured standard output stream.

        :return: string containing the output stream, or `None` if no output was captured
        """
        return None if self._stdout.tell() == 0 else self._stdout.getvalue()

    def get_stderr(self) -> typing.Optional[str]:
        """Return the captured standard error stream.

        :return: string containing the error stream, or `None` if no output was captured
        """
        return None if self._stderr.tell() == 0 else self._stderr.getvalue()


class AnalyticInput:
    """Represents an input to an analytic spec."""

    def __init__(self, name: str, input_type: str, file: str) -> None:
        """Create an input to an analytic spec.

        :param name: the name of the input
        :param input_type: whether the input is a ``table``, a ``column``, or a ``value``
        :param file: the filename of the SBDF file that contains the data to read into this input
        """
        self.name = name
        self.type = input_type
        self.file = file

    def __repr__(self) -> str:
        return f"{_utils.type_name(type(self))}({self.name!r}, {self.type!r}, {self.file!r})"

    def read(self, globals_dict: typing.Dict[str, typing.Any], debug_fn: typing.Callable[[str], None]) -> None:
        """Read an input from the corresponding SBDF file into the dict that comprises the set of globals.

        :param globals_dict: dict containing the global variables for the data function
        :param debug_fn: logging function for debug messages
        """
        if self.type == "NULL":
            debug_fn(f"assigning missing '{self.name}' as None")
            globals_dict[self.name] = None
            return
        debug_fn(f"assigning {self.type} '{self.name}' from file {self.file}")
        dataframe = sbdf.import_data(self.file)
        debug_fn(f"read {dataframe.shape[0]} rows {dataframe.shape[1]} columns")
        try:
            table_meta = dataframe.spotfire_table_metadata
        except AttributeError:
            table_meta = {}
        column_meta = {}
        for col in dataframe.columns:
            try:
                column_meta[col] = dataframe[col].spotfire_column_metadata
            except AttributeError:
                column_meta[col] = {}
        pretty_table = io.StringIO()
        pretty_column = io.StringIO()
        pprint.pprint(table_meta, pretty_table)
        pprint.pprint(column_meta, pretty_column)
        debug_fn(f"table metadata: \n {pretty_table.getvalue()}")
        debug_fn(f"column metadata: \n {pretty_column.getvalue()}")
        if self.type == "column":
            dataframe = dataframe[dataframe.columns[0]]
        if self.type == "value":
            value = dataframe.at[0, dataframe.columns[0]]
            if type(value).__module__ == "numpy":
                dataframe = value.tolist()
            elif type(value).__module__ == "pandas._libs.tslibs.timedeltas":
                dataframe = value.to_pytimedelta()
            elif type(value).__module__ == "pandas._libs.tslibs.timestamps":
                dataframe = value.to_pydatetime()
            elif type(value).__module__ == "pandas._libs.tslibs.nattype":
                dataframe = None
            else:
                dataframe = value
        globals_dict[self.name] = dataframe


class AnalyticOutput:
    """Represents an output from an analytic spec."""

    def __init__(self, name: str, file: str) -> None:
        """Create an output from an analytic spec.

        :param name: the name of the output
        :param file: the filename of the SBDF file that will be created by writing from this output
        """
        self.name = name
        self.file = file

    def __repr__(self) -> str:
        return f"{_utils.type_name(type(self))}({self.name!r}, {self.file!r})"

    def write(self, globals_dict: typing.Dict[str, typing.Any], debug_fn: typing.Callable[[str], None]) -> None:
        """Write an output from the dict that comprises the set of globals file into the corresponding SBDF.

        :param globals_dict: dict containing the global variables from the data function
        :param debug_fn: logging function for debug messages
        """
        debug_fn(f"returning '{self.name}' as file {self.file}")
        sbdf.export_data(globals_dict[self.name], self.file, default_column_name=self.name)


class AnalyticResult:
    """Represents the results of evaluating an AnalyticSpec object."""

    # pylint: disable=too-many-instance-attributes
    # Eight is reasonable in this case.

    def __init__(self) -> None:
        self.success = True
        self.has_stderr = False
        self.std_err_out = None
        self.summary = None
        self._exc_info = (type(None), None, None)
        self._capture = None
        self._debug_log = None

    def fail_with_exception(self, exc_info) -> None:
        """Set this result as failed with an exception."""
        self.fail()
        self._exc_info = exc_info

    def fail(self) -> None:
        """Set the result to Failed."""
        self.success = False

    def set_capture(self, capture: _OutputCapture) -> None:
        """Set the captured standard output and error of this result."""
        self._capture = capture

    def get_capture(self) -> _OutputCapture:
        """Get the captured standard output and error of this result."""
        return self._capture

    def set_debug_log(self, log: str) -> None:
        """Set the debug log that generated this result."""
        self._debug_log = log

    def get_debug_log(self) -> str:
        """Get the debug log for this result."""
        return self._debug_log

    def get_exc_info(self):
        """Get the exception information."""
        return self._exc_info


class AnalyticSpec:
    """Represents an analytic spec used to process a data function."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, analytic_type: str, inputs: typing.List[AnalyticInput], outputs: typing.List[AnalyticOutput],
                 script: str) -> None:
        """Create an analytic spec.

        :param analytic_type: the type of data function being specified; currently only ``script`` is supported
        :param inputs: list of AnalyticInput objects describing inputs to this data function
        :param outputs: list of AnalyticOutput objects describing outputs from this data function
        :param script: the Python script code this data function will run
        """
        self.analytic_type = analytic_type
        self.inputs = inputs if isinstance(inputs, list) else []
        self.outputs = outputs if isinstance(outputs, list) else []
        self.script = script
        self.debug_enabled = False
        self.globals = dict(__builtins__=__builtins__)
        self.log = io.StringIO()
        self.compiled_script = None

    def __repr__(self) -> str:
        return f"{_utils.type_name(type(self))}({self.analytic_type!r}, \
            {self.inputs!r}, {self.outputs!r}, {self.script!r})"

    def enable_debug(self) -> None:
        """Turn on the printing of debugging messages."""
        self.debug_enabled = True

    def debug(self, message: str) -> None:
        """Output a debugging message to the log if debug is enabled.

        :param message: the debugging message to print
        """
        if self.debug_enabled:
            self.log.write(f"debug: {message}\n")

    def debug_write_script(self) -> None:
        """Output the current script to the debugging log."""
        if self.debug_enabled:
            self.log.write(self.script)
            self.log.write("\n")

    def evaluate(self) -> AnalyticResult:
        """Run the script in the in the analytic spec and process the results.

        :return: whether the data function was evaluated successfully, the text output of the data function or an error
        message describing why the data function failed, and the exception information and stack trace if the data
        function raised any errors
        """
        self.debug("start evaluate")
        result = AnalyticResult()

        # noinspection PyBroadException
        try:
            with _OutputCapture() as capture:
                result.set_capture(capture)

                # parse and compile the script first
                self._compile_script(result)
                if not result.success:
                    return self._summarize(result)

                # read inputs
                self._read_inputs(result)
                if not result.success:
                    return self._summarize(result)

                # execute the script
                self._execute_script(self.compiled_script, result)
                if not result.success:
                    return self._summarize(result)

                # write outputs
                self._write_outputs(result)
                if not result.success:
                    return self._summarize(result)

                self.debug("end evaluate")

                result.success = True

            # summarize the evaluation
            self._summarize(result)
        except BaseException:
            self.debug("There was an unhandled exception during evaluation.")
            result.fail_with_exception(sys.exc_info())
            self._summarize(result)
        # return the result object.
        return result

    def _compile_script(self, result: AnalyticResult) -> None:
        """"Parse and compile the script"""
        # noinspection PyBroadException
        try:
            self.compiled_script = compile(self.script, '<data_function>', 'exec')
        except BaseException:
            self.debug("script can't be parsed:")
            self.debug("--- script ---")
            self.debug_write_script()
            self.debug("--- script ---")
            result.fail_with_exception(sys.exc_info())
            return

    def _read_inputs(self, result: AnalyticResult) -> None:
        """read inputs"""
        self.debug(f"reading {len(self.inputs)} input variables")
        for i, input_ in enumerate(self.inputs):
            if _bad_string(input_.name) or input_.name == "":
                self.debug(f"input {i}: bad input variable name - skipped")
                continue
            if input_.type != "NULL" and (_bad_string(input_.file) or input_.file == ""):
                self.debug(f"input '{input_.name}': bad file name - skipped")
                continue
            if input_.type != "NULL" and not os.path.isfile(input_.file):
                self.debug(f"input '{input_.name}': non-existent file - skipped")
                continue
            try:
                input_.read(self.globals, self.debug)
            except sbdf.SBDFError as exc:
                self.debug(f"error reading input variable '{input_.name}' from file '{input_.file}': {str(exc)}")
                result.fail_with_exception(sys.exc_info())
                return
        self.debug(f"done reading {len(self.inputs)} input variables")
        return

    def _execute_script(self, compiled_script: typing.Any, result: AnalyticResult) -> None:
        """execute the script"""
        # pylint: disable=exec-used
        self.debug("executing script")
        self.debug("--- script ---")
        self.debug_write_script()
        self.debug("--- script ---")
        if _bad_string(self.analytic_type):
            self.analytic_type = "script"
        self.debug(f"analytic_type is '{self.analytic_type}'")
        if self.analytic_type == "script":
            # noinspection PyBroadException
            try:
                exec(compiled_script, self.globals)
            except BaseException:
                result.fail_with_exception(sys.exc_info())
                return
        elif self.analytic_type == "aggregationScript":
            self.debug("aggregation scripts are not supported")
            raise DataFunctionError("Aggregation scripts are not supported.")
        self.debug("done executing script")
        return

    def _write_outputs(self, result: AnalyticResult) -> None:
        """write outputs"""
        self.debug(f"writing {len(self.outputs)} output variables")
        for i, output in enumerate(self.outputs):
            if _bad_string(output.name) or output.name == "":
                self.debug(f"output {i}: bad output variable name - skipped")
                continue
            if _bad_string(output.file) or output.file == "":
                self.debug(f"output '{output.name}': bad file name - skipped")
                continue
            if output.name not in self.globals:
                self.debug(f"output variable '{output.name}' not defined in globals")
                raise DataFunctionError(f"Output variable '{output.name}' was not defined")
            # noinspection PyBroadException
            try:
                output.write(self.globals, self.debug)
            except BaseException:
                result.fail_with_exception(sys.exc_info())
        self.debug(f"done writing {len(self.outputs)} output variables")

    def _summarize(self, result: AnalyticResult) -> AnalyticResult:
        """Perform final summary actions and return the result object."""
        result.set_debug_log(self.log.getvalue())
        self._create_summary(result)
        return result

    def _create_summary(self, result: AnalyticResult) -> None:
        """Format the result as a string that can be displayed to the user and provide access to stderr."""
        buf = io.StringIO()
        # noinspection PyBroadException
        # pylint: disable=too-many-nested-blocks
        try:
            if not result.success:
                buf.write("Error executing Python script:\n\n")
                if result.get_exc_info()[2] is not None:
                    # noinspection PyBroadException
                    try:
                        # If it was a syntax error, show the text with the caret
                        exc_type = _utils.type_name(result.get_exc_info()[0])
                        # pylint: disable=consider-using-in
                        if exc_type == "SyntaxError" or exc_type == "IndentationError" or exc_type == "TabError":
                            syntax_error = traceback.TracebackException(result.get_exc_info()[0],
                                                                        result.get_exc_info()[1],
                                                                        result.get_exc_info()[2])
                            buf.write(f'  File "{syntax_error.filename}", line {syntax_error.lineno}\n')
                            buf.write(f"    {syntax_error.text}")
                            # sometimes the offset is wrong, placing the caret before or after the text.
                            spacer = syntax_error.offset - 1
                            if exc_type == "IndentationError":
                                spacer += 1
                            if len(syntax_error.text) == syntax_error.offset:
                                spacer -= 1
                            offset = " " * spacer
                            buf.write(f"    {offset}^\n")
                        buf.write(f"{exc_type}: {result.get_exc_info()[1]}\n\n")
                        buf.write("Traceback (most recent call last):\n")
                        # get the StackSummary
                        tb_frames = traceback.extract_tb(result.get_exc_info()[2])
                        # get the FrameSummary objects from the StackSummary
                        lines = traceback.format_list(tb_frames)

                        # trim the path from the filename
                        def shorten(match):
                            """trim the path from the exception filename"""
                            return f'File "{os.path.basename(match.group(1))}"'

                        lines = [re.sub(r'File "([^"]+)"', shorten, line, 1) for line in lines]
                        buf.writelines(lines)
                    except BaseException:
                        buf.write(f"\nCould not retrieve stacktrace: {traceback.print_exc()}")
            if result.get_capture() is not None:
                out = result.get_capture().get_stdout()
                if out is not None:
                    buf.write("\nStandard output:\n")
                    buf.write(out)
                result.std_err_out = result.get_capture().get_stderr()
                if result.std_err_out is not None:
                    result.has_stderr = True
                    buf.write("\nStandard error:\n")
                    buf.write(result.std_err_out)
            if self.debug_enabled:
                if result.get_debug_log() is not None:
                    buf.write("\nDebug log:\n")
                    buf.write(result.get_debug_log())
        except SystemError as err:
            buf.write(str(err))
        result.summary = buf.getvalue()


# Exceptions

class DataFunctionError(Exception):
    """An exception that is raised to indicate a problem with the Data Function."""
